import json

from bs4 import BeautifulSoup


class SymMessageParser:
    """
    MessageParser class takes message_data as its only parameter.
    message action data is a json object passed from DataFeedEventService
    to the RoomListenerImplementation or IMListenerImplementation class.
    The message_data json object follows the structure below:

    "messageId": "KcIMpk9tHtY8zMmkyGyCK3___pNqVF2QbQ",
    "timestamp": 1565879149167,
    "message": "<div data-format=\"PresentationML\" data-version=\"2.0\" class=\"wysiwyg\"><p><span class=\"entity\" data-entity-id=\"0\">@reed-bot-demo</span> </p></div>",
    "data": "{\"0\":{\"id\":[{\"type\":\"com.symphony.user.userId\",\"value\":\"344147139497165\"}],\"type\":\"com.symphony.user.mention\"}}",
    "user": {
        "userId": 344147139494862,
        "firstName": "Reed",
        "lastName": "Feldman",
        "displayName": "Reed Feldman (SUP)",
        "email": "reed.feldman@symphony.com",
        "username": "reedUAT"
    },
    "stream": {
        "streamId": "pDWC8aUE7IYlK8b9lChaM3___pNuPcIWdA",
        "streamType": "ROOM"
    },
    "externalRecipients": false,
    "userAgent": "DESKTOP-40.0.0-10600-MacOSX-10.14.6-Chrome-76.0.3809.100",
    "originalFormat": "com.symphony.messageml.v2"
}

    This class provides methods to the developer to easily access data in this
    message_data payload for quick bot development.
    """

    def __init__(self):
        self.HASHTAG_TYPE = "org.symphonyoss.taxonomy.hashtag"
        self.CASHTAG_TYPE = "org.symphonyoss.fin.security.id.ticker"
        self.MENTION_TYPE = "com.symphony.user.userId"

    def get_text(self, message_data):
        text_arr = []
        msg_xml = message_data['message']
        soup = BeautifulSoup(msg_xml, 'html.parser')
        for i in soup.find_all('div'):
            text_arr.extend(i.text.split(' '))
        return text_arr

    def get_im_first_name(self, message_data):
        return message_data['user']['firstName']

    def get_im_last_name(self, message_data):
        return message_data['user']['lastName']

    def get_im_name(self, message_data):
        return str(message_data['user']['firstName']) + ' ' + str(message_data['user']['lastName'])

    def get_stream_id(self, message_data):
        return message_data['stream']['streamId']

    def __get_tags(self, json_nodes, tag_type):
        tags = []
        if json_nodes:
            for k, v in json_nodes.items():
                if 'id' in v and len(v['id']) > 0 and 'type' in v['id'][0] and v['id'][0]['type'] == tag_type and 'value' in v['id'][0]:
                    tags.append(v['id'][0]['value'])
        return tags

    def get_mentions(self, message_data):
        mention_arr = []
        msg_xml = message_data['message']
        soup = BeautifulSoup(msg_xml, 'html.parser')
        for i in soup.find_all('span'):
            mention_arr.append(i.text)
        return mention_arr

    def get_mention_ids(self, message_data):
        d = json.loads(message_data['data'])
        return self.__get_tags(d, self.MENTION_TYPE)

    def get_hash_tags(self, message_data):
        tag_arr = []
        msg_xml = message_data['message']
        soup = BeautifulSoup(msg_xml, 'html.parser')
        for i in soup.find_all('span'):
            tag_arr.append(i.text)
        return tag_arr

    def get_hash_tag_values(self, message_data):
        d = json.loads(message_data['data'])
        return self.__get_tags(d, self.HASHTAG_TYPE)

    def get_cash_tags(self, message_data):
        cash_arr = []
        msg_xml = message_data['message']
        soup = BeautifulSoup(msg_xml, 'html.parser')
        for i in soup.find_all('span'):
            cash_arr.append(i.text)
        return cash_arr

    def get_cash_tag_values(self, message_data):
        d = json.loads(message_data['data'])
        return self.__get_tags(d, self.CASHTAG_TYPE)




"""Dummy datafeed services for testing bots

Originally these dummy datafeeds were solely used in the tests, however it was deemed that they might
be of use to downstream clients who want to test their own bots in the same way that they are tested
in the API.

Some of the methods in the SDK use requests, and some use aiohttp, therefore to stub out both of them
two mocking libraries are used. It would have been preferable to mock both types of method with one
library but the only one that offered that possibility was Mocket. Mocket doesn't support regex path
matching and dynamic responses. requests-mock and aioresponses were chosen, the Datafeed mocker wraps
both of these.

To see examples of how things mocks are used, check the tests in test_datafeed.py
"""

import asyncio
import copy
import logging
import re
import uuid
from itertools import count

import requests_mock
from aioresponses import aioresponses, CallbackResult

from sym_api_client_python.datafeed_event_service import DataFeedEventService, AsyncDataFeedEventService

STOP_EVENT = 'TEST_STOP_EVENT'
READ_DATAFEED_URL = re.compile('^.*/agent/v4/datafeed/.*/read')

_CREATE_DATAFEED_URL = '/agent/v4/datafeed/create'
_SESSION_INFO = '/pod/v2/sessioninfo'

BASIC_EVENT = {
    "id": "ulPr8a:eFFDL7",
    "messageId": "1234",
    "timestamp": 1536346282592,
    "type": "MESSAGESENT",
    "initiator": {
        "user": {
            "userId": "1456852...",
            "displayName": "Local Bot01",
            "email": "bot.user1@test.com",
            "username": "bot.user1"
        }
    },
    "payload": {
        "messageSent": {
            "message": {
                "messageId": "CszQa6uPAA9...",
                "timestamp": 1536346282592,
                "message": "<div data-format=\"PresentationML\" data-version=\"2.0\">Hello World</div>",
                "data": "{\"entityIdentifier\":{\"type\":\"org.symphonyoss.fin.security\",\"version\":\"0.1\",\"id\":[{\"type\":\"org.symphonyoss.fin.security.id.isin\",\"value\":\"US0378\"},{\"type\":\"org.symphonyoss.fin.security.id.cusip\",\"value\":\"037\"},{\"type\":\"org.symphonyoss.fin.security.id.openfigi\",\"value\":\"BBG000\"}]}}",
                "user": {
                    "userId": "14568529...",
                    "displayName": "Local Bot01",
                    "email": "bot.user1@ntest.com",
                    "username": "bot.user1"
                },
                "stream": {
                    "streamId": "wTmSDJSNPXgB...",
                    "streamType": "IM"
                },
                "externalRecipients": False,
                "userAgent": "Agent-2.2.8-Linux-4.9.77-31.58.amzn1.x86_64",
                "originalFormat": "com.symphony.messageml.v2"
            }
        }
    }
}


class SymphonyApiMocker:
    """Mocker to wrap requests_mock and aioresponses in one

    For testing the synchronous datafeed, only requests_mock is needed. However for testing the
    async datafeed both are needed as requests_mock handles the mocking for the methods that
    don't have an async counterpart like create_datafeed().

    Pass aio=True to mock through to the asynchronous datafeed, which uses aiohttp for REST
    requests and aioresponses for the mocking.

    Example usage:

        event_service = StoppableService(self.bot_client)
        events = [make_events(count=2)]

        with SymphonyApiMocker() as m:
            m.add_mock_events(events)
            event_service.start_datafeed()

    """

    def __init__(self, aio=False):
        self.requests_mock_manager = requests_mock.Mocker()
        if aio:
            self.aioresponses_manager = aioresponses()
        self.aio = aio

    def __enter__(self):
        """This function just makes sure various core URLs like the create and session datafeed
        examples are allowed to make real http requests. Without this NoMockFound errors will be
        thrown when non-mocked errors are accessed."""
        self.requests_mock_manager.register_uri(
            'POST', _CREATE_DATAFEED_URL, status_code=200, json=self.dummy_id_provider()
        )
        self.requests_mock_manager.register_uri('GET', _SESSION_INFO, real_http=True)
        self.requests_mock_manager.__enter__()
        if self.aio:
            self.aioresponses_manager.__enter__()
        return self

    def __exit__(self, *args):
        self.requests_mock_manager.__exit__(*args)

        if self.aio:
            self.aioresponses_manager.__exit__(*args)

    def add_mock_events(self, events_list, add_stop_event=True):
        """Add a list of events to the mocker. It's expected that most of the time a terminating
        stop event is desired, otherwise the datafeed_read will block forever. If it's not
        required pass add_stop_event=False
        """

        if self.aio:
            # Mock for aioreponses
            for events in events_list:
                self.aioresponses_manager.get(READ_DATAFEED_URL, **events)

            if add_stop_event:
                self.aioresponses_manager.get(
                    READ_DATAFEED_URL, **make_events([STOP_EVENT], aio=True)
                )
                self.aioresponses_manager.get(
                    READ_DATAFEED_URL, callback=make_timed_callback(5), repeat=True
                )
        else:
            if add_stop_event:
                events_list.append(make_events([STOP_EVENT]))
            # anyrequest that goes through this url will return events_list
            # find event list
            self.requests_mock_manager.register_uri('GET', READ_DATAFEED_URL, events_list)

    def dummy_id_provider(self):
        counter = count(1)

        def f(*args, _counter=counter, **kwargs):
            feed_response = {
                "id": "DUMMY-DATAFEED-ID-{}".format(next(_counter))
            }
            return feed_response

        return f


def make_events(event_types=None, count=None, messages=None, aio=False):
    """Returns a dict to be consumed by the SymphonyApiMocker, specifically for the
    read_datafeed API.
        event = make_event()
        m.register_uri('GET', SOME_URL, **event)
    """
    events = []
    if event_types is None:
        count = 1 if count is None else count
        event_types = [None for _ in range(count)]
    else:
        if count is not None and len(event_types) != count:
            raise ValueError("Mismatching length")

    if messages is None:
        messages = [None for _ in event_types]

    for event_type, message in zip(event_types, messages):
        event = copy.deepcopy(BASIC_EVENT)

        # Generate unique event and message ids
        event["id"] = "EVENT-" + str(uuid.uuid4())
        message_id = "MESSAGE-" + str(uuid.uuid4())
        if "messageId" in event:
            event["messageId"] = message_id

        if "messageId" in event["payload"]["messageSent"]["message"]:
            event["payload"]["messageSent"]["message"]["messageId"] = message_id
        else:
            logging.error(event["payload"]["messageSent"]["message"].keys())
        if event_type is not None:
            event['type'] = event_type
        if message is not None:
            original = event["payload"]["messageSent"]["message"]["message"]
            modified = original.replace("Hello World", message)
            event["payload"]["messageSent"]["message"]["message"] = modified
        events.append(event)

    # Different mockers take kwargs in different forms
    if aio:
        return {"payload": events, "status": 200}
    else:
        return {"json": events, "status_code": 200}


def make_error(status_code, error_message=None, aio=False):
    """Returns an error as a dict to be consumed by the SymphonyApiMocker like:
        error = make_error()
        m.register_uri('GET', SOME_URL, **error)
    """
    if error_message is None:
        # Replace with the generic Symphony error message
        error_message = "An error occurred"

    json_body = {"code": status_code, "message": error_message}

    if aio:
        return {"payload": json_body, "status": status_code}
    else:
        return {"json": json_body, "status_code": status_code}


def make_timed_callback(sleep_time=1):
    """Simulate the behaviour of the server, blocking for 30s and return a 204. Instead
    of waiting 30s by default it just waits 1"""

    async def timed_callback(url, **kwargs):
        await asyncio.sleep(sleep_time)
        return CallbackResult(status=204)

    return timed_callback


class StoppableService(DataFeedEventService):
    """The synchronous datafeed blocks on read, meaning anything that tests it will run forever.
    This custom version allows for passing in a STOP event to cancel reading, and end the test.
    All tests that are expected to end without exception should pass a final stop event.
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.baseline_timeout_sec = 2
        self.current_timeout_sec = 2
        self.lower_threshold = 2

        if STOP_EVENT not in self.routing_dict:
            self.routing_dict[STOP_EVENT] = self._stop_feed_handler
        else:
            raise KeyError(f"Dummy stop event {STOP_EVENT} already exists as a regular event")

    def _stop_feed_handler(self, payload):
        self.deactivate_datafeed()


class StoppableAsyncService(AsyncDataFeedEventService):
    """While other strategies could be employed to stop the async services, it's simpler for the
    purposes of these tests to keep the sync model.
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        if STOP_EVENT not in self.routing_dict:
            self.routing_dict[STOP_EVENT] = self._stop_feed_handler
        else:
            raise KeyError(f"Dummy stop event {STOP_EVENT} already exists as a regular event")

    async def _stop_feed_handler(self, payload):
        asyncio.ensure_future(self.deactivate_datafeed())



import os
import unittest
from tests.util.resource_util import get_resource_filepath

from sym_api_client_python.configure.configure import SymConfig

class TestSymConfig(unittest.TestCase):

	def setUp(self):
		self.config = SymConfig(get_resource_filepath('./bot-config.json'))
		self.config.load_config()

	def test_context_path(self):
		self.assertEqual(self.config.data['sessionAuthHost'], "https://MY_ENVIRONMENT.symphony.com:443/sessionAuthContext")
		self.assertEqual(self.config.data['keyAuthHost'], "https://MY_ENVIRONMENT.symphony.com:443/keyAuthContext")
		self.assertEqual(self.config.data['podHost'], "https://MY_ENVIRONMENT.symphony.com:443/podContext")
		self.assertEqual(self.config.data['agentHost'], "https://MY_ENVIRONMENT.symphony.com:443")



import asyncio
import logging
from .room_listener import RoomListener
from .simple_message_processor import MessageProcessor
from sym_api_client_python.processors.sym_message_parser import SymMessageParser
# A sample implementation of Abstract RoomListener class
# The listener can respond to incoming events if the respective event
# handler has been implemented


class RoomListenerTestImp(RoomListener):
    """Example implementation of RoomListener

        sym_bot_client: contains clients which respond to incoming events

    """

    def __init__(self, sym_bot_client):
        self.bot_client = sym_bot_client

    def on_room_msg(self, msg):
        logging.debug('room msg received', msg)
        msg_processor = MessageProcessor(self.bot_client)
        msg_processor.process(msg)

    def on_room_created(self, room_created):
        logging.debug('room created', room_created)

    def on_room_deactivated(self, room_deactivated):
        logging.debug('room Deactivated', room_deactivated)

    def on_room_member_demoted_from_owner(self,
                                          room_member_demoted_from_owner):
        logging.debug('room member demoted from owner',
                      room_member_demoted_from_owner)

    def on_room_member_promoted_to_owner(self, room_member_promoted_to_owner):
        logging.debug('room member promoted to owner',
                      room_member_promoted_to_owner)

    def on_room_reactivated(self, room_reactivated):
        logging.debug('room reactivated', room_reactivated)

    def on_room_updated(self, room_updated):
        logging.debug('room updated', room_updated)

    def on_user_joined_room(self, user_joined_room):
        logging.debug('USER JOINED ROOM', user_joined_room)

    def on_user_left_room(self, user_left_room):
        logging.debug('USER LEFT ROOM', user_left_room)


class AsyncRoomListenerImp(RoomListener):
    """Example implementation of RoomListener with asynchronous functionality
        
    Call the bot with /wait to see an example of a non-blocking wait
    """

    def __init__(self, sym_bot_client):
        self.bot_client = sym_bot_client
        self.message_parser = SymMessageParser()

    async def on_room_msg(self, msg):
        logging.debug('room msg received', msg)

        await asyncio.sleep(5)

        msg_text = self.message_parser.get_text(msg)

        if "/wait" in msg_text:
            await asyncio.sleep(5)
            msg_to_send = dict(
                    message='<messageML>Hello {}, sorry to keep you waiting!</messageML>'.format(
                        self.message_parser.get_im_first_name(msg))
                    )
        else:
            msg_to_send = dict(
                    message='<messageML>Hello {}, hope you are doing well!</messageML>'.format(
                        self.message_parser.get_im_first_name(msg))
                    )

        if msg_text:
            stream_id = self.message_parser.get_stream_id(msg)
            self.bot_client.get_message_client(). \
                    send_msg(stream_id, msg_to_send)


    async def on_room_created(self, room_created):
        logging.debug('room created', room_created)

    async def on_room_deactivated(self, room_deactivated):
        logging.debug('room Deactivated', room_deactivated)

    async def on_room_member_demoted_from_owner(self,
                                          room_member_demoted_from_owner):
        logging.debug('room member demoted from owner',
                      room_member_demoted_from_owner)

    async def on_room_member_promoted_to_owner(self, room_member_promoted_to_owner):
        logging.debug('room member promoted to owner',
                      room_member_promoted_to_owner)

    async def on_room_reactivated(self, room_reactivated):
        logging.debug('room reactivated', room_reactivated)

    async def on_room_updated(self, room_updated):
        logging.debug('room updated', room_updated)

    async def on_user_joined_room(self, user_joined_room):
        logging.debug('USER JOINED ROOM', user_joined_room)

    async def on_user_left_room(self, user_left_room):
        logging.debug('USER LEFT ROOM', user_left_room)



import logging
import time

from .abstract_datafeed_event_service import AbstractDatafeedEventService

from ..exceptions.UnauthorizedException import UnauthorizedException
from ..exceptions.APIClientErrorException import APIClientErrorException
from ..exceptions.DatafeedExpiredException import DatafeedExpiredException
from ..exceptions.ServerErrorException import ServerErrorException
from ..exceptions.MaxRetryException import MaxRetryException

log = logging.getLogger(__name__)


class DataFeedEventServiceV1(AbstractDatafeedEventService):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.datafeed_id = None

    def start_datafeed(self):
        log.debug('DataFeedEventService/startDataFeed()')
        self.datafeed_id = self._get_from_file_or_create_datafeed_id()
        self.read_datafeed()

    def activate_datafeed(self):
        if self.stop:
            self.stop = False

    def deactivate_datafeed(self):
        if not self.stop:
            self.stop = True

    def read_datafeed(self):
        """
            Read_datafeed function reads an array of events coming back from DataFeedClient.

            The json objects returned from read_datafeed() gets passed to handle_events().
        """
        while not self.stop:
            try:
                events = self.datafeed_client.read_datafeed(self.datafeed_id)
            except Exception as exc:
                self.handle_datafeed_errors(exc)
                continue

            self.decrease_timeout()
            if events:
                self.handle_events(events)
            else:
                log.debug(
                    'DataFeedEventService() - no data coming in from '
                    'datafeed: {}'.format(self.datafeed_id)
                )

    ### Handling errors ###
    def handle_datafeed_errors(self, thrown_exception):
        """Various errors may get thrown by the datafeed reader, from 500s when a server node is
        being bounced, intermittent connectivity or SSL problems. With the exception of a 403
        UnauthorisedError where reauthentication occurs upstream, these are all handled in
        the same way - sleep, request a new datafeed id and then retry.
        """
        try:
            raise thrown_exception
        except UnauthorizedException:
            log.error('DataFeedEventService - caught unauthorized exception')
        except MaxRetryException as e:
            log.error('DataFeedEventService - Bot has tried to authenticate more than 5 times ')
            raise

        except (DatafeedExpiredException, APIClientErrorException, ServerErrorException) as exc:
            log.error('DataFeedEventService - ' + str(exc))
        except Exception as exc:
            log.exception('DataFeedEventService - Unknown exception: ' + str(exc))
        sleep_for = self.get_and_increase_timeout(thrown_exception)
        log.debug('DataFeedEventService/handle_event() --> Sleeping for {:.4g}s'.format(sleep_for))
        time.sleep(sleep_for)

        try:
            log.debug('DataFeedEventService/handle_event() --> Restarting Datafeed')
            self.datafeed_id = self._create_datafeed_and_persist()

        except Exception as exc:
            self.handle_datafeed_errors(exc)



from abc import ABC, abstractmethod

# Abstract class for wallpost listener
# class is just an interface of functions to handle the connection events
# received from DataFeed Connection Events are "SHAREDPOST" and
# "MESSAGESENT" handle actual event logic in your implementation of
# this abstract class


class WallPostListener(ABC):

    @abstractmethod
    def on_wall_post_msg(self, user):
        pass

    @abstractmethod
    def on_shared_post(self, user):
        pass



import logging

from .api_client import APIClient

# child class of APIClient --> Extends error handling functionality
# MessageClient class contains a series of functions corresponding to all
# messaging endpoints on the REST API.

MESSAGE_STREAM_API = '/agent/v4/stream/{stream_id}/message'
MESSAGE_CREATE = MESSAGE_STREAM_API + '/create'


class MessageClient(APIClient):

    def __init__(self, bot_client):
        self.bot_client = bot_client

    def get_msg_from_stream(self, stream_id, since, **kwargs):
        logging.debug('MessageClient/get_msg_from_stream()')
        url = MESSAGE_STREAM_API.format(stream_id=stream_id)
        params = {
            'since': since
        }
        params.update(kwargs)
        return self.bot_client.execute_rest_call('GET', url, params=params)

    async def get_msg_from_stream_async(self, stream_id, since, **kwargs):
        logging.debug('MessageClient/get_msg_from_stream_async()')
        url = MESSAGE_STREAM_API.format(stream_id=stream_id)
        params = {
            'since': since
        }
        params.update(kwargs)
        return await self.bot_client.execute_rest_call_async('GET', url, params=params)

    def send_msg(self, stream_id, outbound_msg):
        logging.debug('MessageClient/send_msg()')
        url = MESSAGE_CREATE.format(stream_id=stream_id)
        return self.bot_client.execute_rest_call('POST', url, files=outbound_msg)

    async def send_msg_async(self, stream_id, outbound_msg):
        logging.debug('MessageClient/send_msg()')
        url = MESSAGE_CREATE.format(stream_id=stream_id)
        return await self.bot_client.execute_rest_call_async('POST', url, files=outbound_msg)

    def _data_and_headers_for_attachment(self, stream_id, msg, filename, path_to_attachment, aio=False):
        """Build an attachment out of either a path or a stream"""
        url = MESSAGE_CREATE.format(stream_id=stream_id)

        try:
            attachment_file_object = open(path_to_attachment, 'rb')
        except TypeError:
            # If it doesn't support open treat it as a stream already
            attachment_file_object = path_to_attachment

        # The below states that Content-Type for attachments should be 'file' which is almost
        # certainly wrong - it's not a valid MIME-type. text/plain seems right
        fields = {'message': msg,
                  'attachment': (
                      filename, attachment_file_object, "file")}

        parts = self.make_mulitpart_form(fields, aio=aio)

        return {'path': url, **parts}

    def send_msg_with_attachment(self, stream_id, msg,
                                 filename, path_to_attachment):
        """In this function make sure that msg parameter is set to just the messageML string.
        Do not set msg parameter to dict(message='<messageML>testing attachement</messageML>')
        for this function"""
        logging.debug('MessageClient/send_msg_with_attachment()')
        parts = self._data_and_headers_for_attachment(stream_id, msg, filename, path_to_attachment)
        return self.bot_client.execute_rest_call("POST", **parts)

    async def send_msg_with_attachment_async(self, stream_id, msg,
                                             filename, path_to_attachment):
        logging.debug('MessageClient/send_msg_with_attachment()')

        parts = self._data_and_headers_for_attachment(stream_id, msg, filename, path_to_attachment, aio=True)
        return await self.bot_client.execute_rest_call_async('POST', **parts)

    def get_msg_attachment(self, stream_id, msg_id, file_id):
        logging.debug('MessageClient/get_msg_attachment()')
        url = '/agent/v1/stream/{0}/attachment'.format(stream_id)
        params = {
            'messageId': msg_id,
            'fileId': file_id
        }
        return self.bot_client.execute_rest_call("GET", url, params=params)

    # go on admin clients --> Contains sample data just for example's sake
    def import_message(self, importedMessage):
        logging.debug('MessageClient/import_message()')
        url = '/agent/v4/message/import'
        return self.bot_client.execute_rest_call("POST", url, json=importedMessage)

    # go on admin clients
    def suppress_message(self, id):
        logging.debug('MessageClient/suppress_message()')
        url = '/pod/v1/admin/messagesuppression/{0}/suppress'.format(id)
        return self.bot_client.execute_rest_call("POST", url)

    def post_msg_search(self, query, **kwargs):
        logging.debug('MessageClient/post_msg_search()')
        url = '/agent/v1/message/search'
        return self.bot_client.execute_rest_call("POST", url, json=query, params=kwargs)

    # contains sample query for example
    def get_msg_search(self, query, **kwargs):
        logging.debug('MessageClient/get_msg_search()')
        url = '/agent/v1/message/search'
        params = {
            'query': query
        }
        params.update(kwargs)
        return self.bot_client.execute_rest_call("GET", url, params=params)

    def get_msg_status(self, msg_id):
        logging.debug('MessageClient/get_msg_status()')
        url = '/pod/v1/message/{0}/status'.format(msg_id)
        return self.bot_client.execute_rest_call("GET", url)

    def get_supported_attachment_types(self):
        logging.debug('MessageClient/getAttachmentTypes()')
        url = '/pod/v1/files/allowedTypes'
        return self.bot_client.execute_rest_call("GET", url)

    def get_msg_ids_by_timestamp(self, msg_id, **kwargs):
        logging.debug('MessageClient/get_msg_ids_by_timestamp()')
        url = '/pod/v2/admin/streams/{0}/messageIds'.format(msg_id)
        return self.bot_client.execute_rest_call('GET', url, params=kwargs)

    def list_msg_receipts(self, msg_id):
        logging.debug('MessageClient/list_msg_receipts()')
        url = '/pod/v1/admin/messages/{0}/receipts'.format(msg_id)
        return self.bot_client.example('GET', url)

    def list_stream_attachments(self, stream_id):
        logging.debug('MessageClient/list_msg_attachments()')
        url = '/pod/v1/streams/{0}/attachments'.format(stream_id)
        return self.bot_client.execute_rest_call('GET', url)



from abc import ABC, abstractmethod

# Abstract class for Elements Action listener
# class is just an interface of functions to handle the Elements Actions events received
# from DataFeed
# (see Real Time Events in REST API documentation for more details)
# the developer will handle actual event logic in your implementation of
# this abstract class


class ElementsActionListener(ABC):

    @abstractmethod
    def on_elements_action(self, action):
        """
        Do Something
        """



import asyncio
import logging
from .im_listener import IMListener
from .simple_message_processor import MessageProcessor
from sym_api_client_python.processors.sym_message_parser import SymMessageParser




# A sample implementation of Abstract imListener class
# The listener can respond to incoming events if the respective event
# handler has been implemented


class IMListenerTestImp(IMListener):
    """Example implementation of IMListener

        sym_bot_client: contains clients which respond to incoming events

    """

    def __init__(self, sym_bot_client):
        self.bot_client = sym_bot_client

    def on_im_message(self, im_message):
        logging.debug('message received in IM')
        msg_processor = MessageProcessor(self.bot_client)
        msg_processor.process(im_message)

    def on_im_created(self, im_created):
        logging.debug('IM created!', im_created)

class AsyncIMListenerImp(IMListener):
    """Example implementation of IMListener with asynchronous functionality

    Call the bot with /wait to see an example of a non-blocking wait

    """

    def __init__(self, sym_bot_client):
        self.bot_client = sym_bot_client
        self.message_parser = SymMessageParser()

    async def on_im_message(self, msg):
        logging.debug('message received in IM', msg)


        msg_text = self.message_parser.get_text(msg)

        if "/wait" in msg_text:
            await asyncio.sleep(5)
            msg_to_send = dict(
                    message='<messageML>Hello {}, sorry to keep you waiting!</messageML>'.format(
                        self.message_parser.get_im_first_name(msg))
                    )
        else:
            msg_to_send = dict(
                    message='<messageML>Hello {}, hope you are doing well!</messageML>'.format(
                        self.message_parser.get_im_first_name(msg))
                    )

        if msg_text:
            stream_id = self.message_parser.get_stream_id(msg)
            self.bot_client.get_message_client(). \
                    send_msg(stream_id, msg_to_send)


    async def on_im_created(self, im_created):
        logging.debug("IM created!", im_created)



import logging

from .api_client import APIClient


# child class of APIClient --> Extends error handling functionality
# SignalsClient class contains a series of functions corresponding to all
# pod admin endpoints on the REST API.
class SignalsClient(APIClient):

    def __init__(self, bot_client):
        self.bot_client = bot_client

    def list_signals(self, skip=0, limit=50):
        """
        Lists signals on behalf of the user. 
        The response includes signals that the user has created 
        and public signals to which they have subscribed.
        """
        logging.debug('SignalsClient/list_signals()')
        url = '/agent/v1/signals/list' 
        params = {'skip': skip, 'limit': limit}
        return self.bot_client.execute_rest_call("GET", url, params=params)

    def get_signal(self, signal_id):
        """
        Gets details about the specified signal.
        """
        logging.debug('SignalsClient/get_signal()')
        url = '/agent/v1/signals/{0}/get'.format(signal_id)
        return self.bot_client.execute_rest_call("GET", url)
    
    def create_signal(self, signal_object):
        """
        Creates a new Signal.

        Reference for the Signal Object:
        https://developers.symphony.com/restapi/reference#signal-object

        To create a company-wide signal, the requesting user needs to have the canCreatePushedSignals entitlement.

        To send numeric cashtags as signals, add a * before the number, for example, $*122450.
        """
        logging.debug('SignalClient/create_signal()')
        url = '/agent/v1/signals/create'
        return self.bot_client.execute_rest_call("POST", url, json=signal_object)
    
    def update_signal(self, signal_id, signal_object):
        """
        Updates an existing Signal.
        
        Reference for the Signal Object:
        https://developers.symphony.com/restapi/reference#signal-object

        To update a company-wide signal, the requesting user needs to have the canCreatePushedSignals entitlement.

        To update a normal signal, the requesting user needs to be the owner of the signal.

        To send numeric cashtags as signals, add a * before the number, for example, $*122450.
        """
        logging.debug('SignalsClient/update_signal()')
        url = '/agent/v1/signals/{0}/update'.format(signal_id)
        return self.bot_client.execute_rest_call("POST", url, json=signal_object)

    def delete_signal(self, signal_id):
        """Deletes an existing Signal."""
        logging.debug('SignalsClient/delete_signal()')
        url = '/agent/v1/signals/{0}/delete'.format(signal_id)
        return self.bot_client.execute_rest_call("POST", url)    

    def subscribe_signal(self, user_id_array, signal_id, pushed=False):
        """
        Subscribe an array of users to a Signal.
        
        To subscribe an entire pod to a Signal, set the companyWide field in Create Signal.

        To subscribe other users to a specific signal, 
        the requesting user needs to have the canManageSignalSubscription entitlement.
        """
        logging.debug('SignalsClient/subscribe_signal()')
        url = '/agent/v1/signals/{0}/subscribe'.format(signal_id)
        params = {"pushed": pushed}
        return self.bot_client.execute_rest_call("POST", url, params=params, json=user_id_array)

    
    def unsubscribe_signal(self, user_id_array, signal_id):
        """
        Unsubscribe an array of users from the specified Signal.
        
        To unsubscribe from a signal, the requesting user cannot be the owner of the signal.   

        To unsubscribe other users from a signal, the requesting user needs to have the canManageSignalSubscription entitlement and the signal cannot be a company-wide signal.
        """
        logging.debug('SignalsClient/unsubscribe_signal()')
        url = '/agent/v1/signals/{0}/unsubscribe'.format(signal_id)
        return self.bot_client.execute_rest_call("POST", url, json=user_id_array)


    def get_subscribers(self, signal_id, skip=0, limit=50):
        """ Gets the subscribers for the specified signal."""
        logging.debug('SignalsClient/get_subscribers()')
        url = '/agent/v1/signals/{0}/subscribers'.format(signal_id) 
        params = {'skip': skip, 'limit': limit}
        return self.bot_client.execute_rest_call("GET", url, params=params)



import json
import requests
import datetime
import time
import logging
from .auth_endpoint_constants import auth_endpoint_constants
from jose import jwt
from ..clients.api_client import APIClient
from ..exceptions.MaxRetryException import MaxRetryException

class SymBotRSAAuth(APIClient):
    """Class for RSA authentication"""

    def __init__(self, config):
        """
        Set up proxy information if configuration contains proxyURL
        :param config: Object contains all RSA configurations
        """
        self.config = config
        self.last_auth_time = 0
        self.auth_retries = 0
        self.session_token = None
        self.key_manager_token = None
        self.auth_session = requests.Session()
        self.key_manager_auth_session = requests.Session()

        self.auth_session.proxies.update(self.config.data['podProxyRequestObject'])
        self.key_manager_auth_session.proxies.update(self.config.data['keyManagerProxyRequestObject'])

        if self.config.data['truststorePath']:
            logging.debug('truststore being added to requests library')
            self.auth_session.verify = self.config.data['truststorePath']
            self.key_manager_auth_session.verify = self.config.data['truststorePath']

    def get_session_token(self):
        """Return the session token"""
        return self.session_token

    def get_key_manager_token(self):
        """Return the key manager token"""
        return self.key_manager_token

    def authenticate(self):
        """
        Get the session and key manager token
        """
        logging.debug('RSA Auth/authenticate()')
        try:
            if (self.last_auth_time == 0) or \
                    (int(round(time.time() * 1000) - self.last_auth_time >= auth_endpoint_constants['WAIT_TIME'])):
                logging.debug('RSA Auth/authenticate() --> needed to authenticate')

                self.last_auth_time = int(round(time.time() * 1000))
                self.session_authenticate()
                self.key_manager_authenticate()

            else:
                logging.debug('Retry authentication in 30 seconds.')
                time.sleep(auth_endpoint_constants['TIMEOUT'])
                self.authenticate()
        except MaxRetryException as e:
            logging.exception(e)
            raise MaxRetryException

    def create_jwt(self):
        """
        Create a jwt token with payload dictionary. Encode with
        RSA private key using RS512 algorithm

        :return: A jwt token valid for < 290 seconds
        """
        logging.debug('RSA_auth/getJWT() function started')
        with open(self.config.data['botRSAPath'], 'r') as f:
            content = f.readlines()
            private_key = ''.join(content)
            expiration_date = int(datetime.datetime.now(datetime.timezone.utc)
                                  .timestamp() + (5*58))
            payload = {
                'sub': self.config.data['botUsername'],
                'exp': expiration_date
            }
            encoded = jwt.encode(payload, private_key, algorithm='RS512')
            f.close()
            return encoded

    def session_authenticate(self):
        """
        Get the session token by calling API using jwt token
        """
        logging.debug('RSA_auth/get_session_token() function started')
        data = {
            'token': self.create_jwt()
        }
        url = self.config.data['sessionAuthUrl']+'/login/pubkey/authenticate'
        response = self.auth_session.post(url, json=data)

        if response.status_code != 200:
            self.auth_retries += 1
            if self.auth_retries > auth_endpoint_constants['MAX_RSA_RETRY']:
                # raise UnauthorizedException('max auth retry limit: {}'.format(response.__dict__))
                raise MaxRetryException('bot failed to authenticate more than 5 times.')
            else:
                logging.debug('RSA_auth/get_session_token() function failed: {}'.format(
                    response.status_code)
                )
                time.sleep(auth_endpoint_constants['TIMEOUT'])
                self.session_authenticate()
        else:
            data = json.loads(response.text)
            logging.debug('RSA/session token success')
            self.session_token = data['token']
            self.auth_retries = 0


    def key_manager_authenticate(self):
        """
        Get the key manager token by calling API using jwt token
        """
        logging.debug('RSA_auth/get_keyauth()')
        data = {
            'token': self.create_jwt()
        }
        url = self.config.data['keyAuthUrl']+'/relay/pubkey/authenticate'
        response = self.key_manager_auth_session.post(url, json=data)

        if response.status_code != 200:
            self.auth_retries += 1
            if self.auth_retries > auth_endpoint_constants['MAX_RSA_RETRY']:

                raise MaxRetryException('bot failed to authenticate more than 5 times.')
            else:
                logging.debug('RSA_auth/get_key_manager_authenticate() function failed: {}'.format(
                    response.status_code)
                )
                time.sleep(auth_endpoint_constants['TIMEOUT'])
                self.key_manager_authenticate()
        else:
            data = json.loads(response.text)
            logging.debug('RSA/key manager token success')
            self.key_manager_token = data['token']
            self.auth_retries = 0



import json
import os
import unittest
from unittest.mock import patch

from sym_api_client_python.auth.rsa_auth import SymBotRSAAuth
from sym_api_client_python.clients.sym_bot_client import SymBotClient
from sym_api_client_python.configure.configure import SymConfig


@patch('sym_api_client_python.clients.sym_bot_client.requests.sessions.Session.request')
class TestDataFeedClientV2(unittest.TestCase):
    def setUp(self):
        configure = SymConfig(get_path_relative_to_resources_folder('./bot-config.json'))
        configure.load_config()
        configure.data['datafeedVersion'] = 'v2'

        bot_client = SymBotClient(SymBotRSAAuth(configure), configure)

        self.datafeed_client = bot_client.get_datafeed_client()

    def test_create_datafeed(self, mock_request):
        mock_response, url_call = mocked_response('CREATE_DATAFEED', self.datafeed_client.config.data['agentUrl'])
        mock_request.return_value = mock_response
        datafeed_id = self.datafeed_client.create_datafeed()

        self.assertEqual(mock_response.status_code, 201)
        self.assertEqual(datafeed_id, '21449143d35a86461e254d28697214b4_f')
        mock_request.assert_called_with('POST', url_call)

    def test_list_datafeed(self, mock_request):
        mock_response, url_call = mocked_response('LIST_DATAFEED', self.datafeed_client.config.data['agentUrl'])
        mock_request.return_value = mock_response
        datafeed_ids = self.datafeed_client.list_datafeed_id()

        self.assertEqual(mock_response.status_code, 200)
        self.assertEqual(datafeed_ids[0]['id'], mock_response.get_json()[0]['id'])
        self.assertEqual(datafeed_ids[1]['id'], mock_response.get_json()[1]['id'])
        self.assertEqual(datafeed_ids[2]['id'], mock_response.get_json()[2]['id'])
        mock_request.assert_called_with('GET', url_call)

    def test_read_datafeed_empty_ackid(self, mock_request):
        """Test Datafeed Read first call conversation

        Test of handling of ackId when not provided to read_datafeed and call of the function with the right parameters"""
        mock_response, url_call = mocked_response('READ_DATAFEED', self.datafeed_client.config.data['agentUrl'],
                                                  'test_datafeed_id')
        mock_request.return_value = mock_response
        events = self.datafeed_client.read_datafeed('test_datafeed_id')
        ack_id = self.datafeed_client.datafeed_client.get_ack_id()

        self.assertEqual(mock_response.status_code, 200)
        self.assertEqual(ack_id, mock_response.get_json()['ackId'])
        self.assertEqual(events, mock_response.get_json()['events'])
        mock_request.assert_called_with('POST', url_call, json={'ackId': ''})

    def test_read_datafeed(self, mock_request):
        """Test a datafeed read during conversation

        Testing the handling of the ackId when provided to read_datafeed() and call of the function with the right parameters """
        mock_response, url_call = mocked_response('READ_DATAFEED', self.datafeed_client.config.data['agentUrl'],
                                                  'test_datafeed_id')
        mock_request.return_value = mock_response
        events = self.datafeed_client.read_datafeed('test_datafeed_id', 'test_ack_id')
        ack_id = self.datafeed_client.datafeed_client.get_ack_id()

        self.assertEqual(mock_response.status_code, 200)
        self.assertEqual(ack_id, mock_response.get_json()['ackId'])
        self.assertEqual(events, mock_response.get_json()['events'])
        mock_request.assert_called_with('POST', url_call, json={'ackId': 'test_ack_id'})

    def test_delete_datafeed(self, mock_request):
        """Test deleting the datafeed

        Because it's a no content response, we need to make sure of the call parameters """
        mock_response, url_call = mocked_response('DELETE_DATAFEED', self.datafeed_client.config.data['agentUrl'],
                                                  'test_datafeed_id')
        mock_request.return_value = mock_response
        self.datafeed_client.delete_datafeed('test_datafeed_id')

        self.assertEqual(mock_response.status_code, 204)
        mock_request.assert_called_with('DELETE', url_call)


def get_path_relative_to_resources_folder(path_relative_to_resources):
    path_to_resources = os.path.join(os.path.dirname(__file__), '../../resources/', path_relative_to_resources)
    return os.path.normpath(path_to_resources)


def url_method_builder(base_url, url):
    return base_url + url


# This method will be used by the mock to replace Session.request
def mocked_response(agent_call, agent_host, *datafeed_id):
    if agent_call == 'CREATE_DATAFEED':
        return MockResponse(201, get_path_relative_to_resources_folder('./response_content/datafeed_v2/create_datafeed_v2.json')), \
               url_method_builder(agent_host, '/agent/v5/datafeeds')
    elif agent_call == 'LIST_DATAFEED':
        return MockResponse(200, get_path_relative_to_resources_folder('./response_content/datafeed_v2/list_datafeed_v2.json')), \
               url_method_builder(agent_host, '/agent/v5/datafeeds')
    elif agent_call == 'READ_DATAFEED':
        if len(datafeed_id) == 0:
            raise ValueError('If datafeed id is provided, it should not be empty')
        else:
            return MockResponse(200, get_path_relative_to_resources_folder('./response_content/datafeed_v2/read_datafeed_v2.json')), \
                   url_method_builder(agent_host, '/agent/v5/datafeeds/{0}/read'.format(datafeed_id[0]))
    elif agent_call == 'DELETE_DATAFEED':
        if len(datafeed_id) == 0:
            raise ValueError('If datafeed id is provided, it should not be empty')
        else:
            return MockResponse(204, ''), url_method_builder(agent_host,
                                                             '/agent/v5/datafeeds/{0}'.format(datafeed_id[0]))
    else:
        return MockResponse(None, 404)


class MockResponse:
    def __init__(self, status_code, path):
        self.status_code = status_code
        if len(path) == 0:
            self.data = ''
        else:
            with open(os.path.realpath(path)) as json_file:
                self.data = json.load(json_file)
        self.text = json.dumps(self.data)

    def get_json(self):
        return self.data

    def get_text(self):
        return json.dumps(self.data)



from enum import Enum
import json

class DatafeedVersion(str, Enum):
    V1: str = "v1"
    V2: str = "v2"

    @classmethod
    def version_of(cls, version):
        if version.lower() == cls.V2.value:
            return cls.V2
        else:
            return cls.V1





from .suppression_listener import SuppressionListener
import logging
# A sample implementation of Abstract SuppressionListener class
# The listener can respond to incoming events if the respective event
# handler has been implemented


class SuppressionListenerImp(SuppressionListener):
    """Example implementation of Message Suppression Listener

        sym_bot_client: contains clients which respond to incoming events

    """

    def __init__(self, sym_bot_client):
        self.bot_client = sym_bot_client

    def on_message_suppression(self, suppressed_message):
        logging.debug('message suppression detected %s', suppressed_message)



import logging
import os

DATAFEED_ID_FILE = 'datafeed.id'

log = logging.getLogger(__name__)


class OnDiskDatafeedIdRepository:
    def __init__(self, datafeed_id_folder):
        self.datafeed_id_file_path = self._get_datafeed_id_file_path(datafeed_id_folder)

    def read_datafeed_id_from_file(self):
        log.debug(f'Retrieving datafeed id from file {self.datafeed_id_file_path}')
        if os.path.exists(self.datafeed_id_file_path):
            with open(self.datafeed_id_file_path, 'r') as datafeed_id_file:
                line = datafeed_id_file.readline()
                if line:
                    index = line.find("@")
                    if index != -1:
                        datafeed_id = line[0:index]
                        log.debug(f'Retrieved datafeed id: {datafeed_id}')
                        return datafeed_id
        log.debug(f'Could not retrieve datafeed id from file {self.datafeed_id_file_path}')
        return None

    def store_datafeed_id_to_file(self, datafeed_id, agent_url):
        with open(self.datafeed_id_file_path, 'w') as datafeed_id_file:
            line = f'{datafeed_id}@{agent_url}'
            log.debug(f'Writing {line} to {self.datafeed_id_file_path}')
            datafeed_id_file.write(line)

    def _get_datafeed_id_file_path(self, datafeed_id_folder):
        datafeed_id_file_path = os.path.join(datafeed_id_folder, DATAFEED_ID_FILE)
        if os.path.exists(datafeed_id_file_path) and os.path.isdir(datafeed_id_file_path):
            datafeed_id_file_path = os.path.join(datafeed_id_file_path, DATAFEED_ID_FILE)
        return os.path.abspath(datafeed_id_file_path)



import logging
import json

from .abstract_datafeed_event_service import AbstractDatafeedEventService

from ..exceptions.UnauthorizedException import UnauthorizedException
from ..exceptions.DatafeedExpiredException import DatafeedExpiredException
from ..exceptions.ServerErrorException import ServerErrorException
from ..exceptions.MaxRetryException import MaxRetryException
from ..exceptions.APIClientErrorException import APIClientErrorException
import time

log = logging.getLogger(__name__)

class DataFeedEventServiceV2(AbstractDatafeedEventService):
    def __init__(self, *args, **kwargs):
        self.datafeed_id = None
        super().__init__(*args, **kwargs)



    def start_datafeed(self):
        log.debug('DataFeedEventServiceV2/startDataFeed()')
        self.read_datafeed()

    def activate_datafeed(self):
        if self.stop:
            self.stop = False

    def deactivate_datafeed(self):
        if not self.stop:
            self.stop = True

    def read_datafeed(self):
        """
            Read_datafeed function reads an array of events coming back from DataFeedClient.

            The json objects returned from read_datafeed() gets passed to handle_events().
        """
        datafeed_ids = self.datafeed_client.list_datafeed_id()

        if len(datafeed_ids) == 0:
            self.datafeed_id = self.datafeed_client.create_datafeed()
        else:
            self.datafeed_id = datafeed_ids[0].get("id")

        while not self.stop:
            try:
                events = self.datafeed_client.read_datafeed(self.datafeed_id, self.datafeed_client.get_ack_id())
            except Exception as exc:
                self.handle_datafeed_errors(exc)
                continue

            self.decrease_timeout()

            if events and events != [None]:
                self.handle_events(events)
            else:
                log.debug(
                    'DataFeedEventServiceV2() - no data coming in from '
                    'datafeed: {}'.format(self.datafeed_id)
                )


    def handle_datafeed_errors(self, thrown_exception):
        """Various errors may get thrown by the datafeed reader, from 500s when a server node is
        being bounced, intermittent connectivity or SSL problems. With the exception of a 403
        UnauthorisedError where reauthentication occurs upstream, these are all handled in
        the same way - sleep, request a new datafeed id and then retry.
        """
        try:
            raise thrown_exception
        except UnauthorizedException:
            log.error('DataFeedEventService - caught unauthorized exception')
        except MaxRetryException as e:
            log.error('DataFeedEventService - Bot has tried to authenticate more than 5 times ')
            raise

        except (DatafeedExpiredException, APIClientErrorException, ServerErrorException) as exc:
            log.error('DataFeedEventService - ' + str(exc))
        except Exception as exc:
            log.exception('DataFeedEventService - Unknown exception: ' + str(exc))
        sleep_for = self.get_and_increase_timeout(thrown_exception)
        log.debug('DataFeedEventService/handle_event() --> Sleeping for {:.4g}s'.format(sleep_for))
        time.sleep(sleep_for)

        try:
            log.debug('DataFeedEventServiceV2 --> Deleting previous Datafeed')
            self.datafeed_client.delete_datafeed(self.datafeed_id)
        except Exception as exc:
            self.handle_datafeed_errors(exc)

        try:
            log.debug('DataFeedEventServiceV2/handle_event() --> Restarting Datafeed')
            self.datafeed_id = self.datafeed_client.create_datafeed()

        except Exception as exc:
            self.handle_datafeed_errors(exc)


import logging

from .api_client import APIClient


# child class of APIClient --> Extends error handling functionality
# StreamClient class contains a series of functions corresponding to all stream
# endpoints on the REST API.


class StreamClient(APIClient):

    def __init__(self, bot_client):
        self.bot_client = bot_client

    def create_im(self, users_array):
        """
        Creates a new single or multi-party instant message conversation or returns
        an existing IM or MIM between the specified users and the calling user.
        """
        logging.debug('StreamClient/create_im()')
        url = '/pod/v1/im/create'
        return self.bot_client.execute_rest_call("POST", url, json=users_array)

    def create_im_admin(self, users_array):
        """
        Creates a new single or multi-party instant message conversation or returns
        an existing IM or MIM between the specified users, but excludes the calling user.
        """
        logging.debug('StreamClient/create_im_admin()')
        url = '/pod/v1/admin/im/create'
        return self.bot_client.execute_rest_call("POST", url, json=users_array)

    def create_room(self, roomToCreate):
        """
        Creates a new chatroom. See Room Attributes for room creation parameters.
        Example Room Object. All required

        roomToCreate = {
            "name": (string) Room name,
            "description": (string) Room description,
            "keywords": [
                {"key": "key1", "value": "value1"},
                {"key": "key2", "value": "value2"}
                ...
            ],
            "membersCanInvite": (bool) If true, any chat room participant can add new participants. If false, only owners can add new participants,
            "discoverable": (bool) If true, this chat room (name, description and messages) non-participants can search for this room. If false, only participants can search for this room,
            "public": (bool) If true, this is a public chatroom. If false, a private chatroom. Note: Once this value is set for a room, it is read-only and can’t be updated,
            "readOnly": (bool) If true, only stream owners can send messages. Once this value is set for a room, it is read-only and can’t be updated,
            "copyProtected": (bool) If true, users cannot copy content from this room. Once this value is set to true for a room, it is read-only and can’t be updated,
            "crossPod": (bool) If true, this room is a cross-pod room,
            "viewHistory": (bool) If true, new members can view the room chat history of the room,
            "multiLateralRoom": (bool) If true, this is a multilateral room where users belonging to more than 2 companies can be found
        }
        """
        logging.debug('StreamClient/create_room()')
        url = '/pod/v3/room/create'
        return self.bot_client.execute_rest_call("POST", url, json=roomToCreate)

    def update_room(self, stream_id, **kwargs):
        """
        Updates the attributes of an existing chat room.
        Pass in room_obj like such:

        room_obj = {
            "name" : "updated room",
            "description": "testing update room function",
        }

        update_room(stream_id, **room_obj)
        """
        logging.debug('StreamClient/update_room()')
        url = '/pod/v3/room/{0}/update'.format(stream_id)
        return self.bot_client.execute_rest_call('POST', url, json=kwargs)

    def get_room_info(self, stream_id):
        """
        Returns information about a particular chat room.
        """
        logging.debug('StreamClient/get_room_info()')
        url = '/pod/v3/room/{0}/info'.format(stream_id)
        return self.bot_client.execute_rest_call('GET', url)

    def activate_room(self, stream_id):
        """
        Deactivate or reactivate a chatroom. At creation, a new chatroom is active.
        """
        logging.debug('StreamClient/activate_room()')
        url = '/pod/v1/room/{0}/setActive'.format(stream_id)
        params = {
            'active': True
        }
        return self.bot_client.execute_rest_call('POST', url, params=params)

    def deactivate_room(self, stream_id):
        """
        Deactivate or reactivate a chatroom. At creation, a new chatroom is active.
        """
        logging.debug('StreamClient/activate_room()')
        url = '/pod/v1/room/{0}/setActive'.format(stream_id)
        params = {
            'active': False
        }
        return self.bot_client.execute_rest_call('POST', url, params=params)

    def get_room_members(self, stream_id):
        """
        Lists the current members of an existing room.
        """
        logging.debug('StreamClient/get_room_members()')
        url = '/pod/v2/room/{0}/membership/list'.format(stream_id)
        return self.bot_client.execute_rest_call('GET', url)

    def add_member_to_room(self, stream_id, user_id):
        """
        Adds a new member to an existing room.
        """
        logging.debug('StreamClient/add_member_to_room()')
        url = '/pod/v1/room/{0}/membership/add'.format(stream_id)
        data = {'id': user_id}
        return self.bot_client.execute_rest_call('POST', url, json=data)

    # Content Object. * is required. Either articleId or articleUrl must be specified
    # "content":{
    #     "articleId": (string) A unique ID for this article, not used by any other article,
    #     "title": (string) The title of the article,
    #     "subTitle": (string) The subtitle of the article
    #     "description": (string) The description of the article,
    #     "message" : (string) The message text that can be sent along with the shared article,
    #     "publisher": (string)* Publisher of the article,
    #     "publishDate": (string) Article publish date in unix timestamp,
    #     "thumbnailUrl": (string) URL to the thumbnail image,
    #     "author": (string)* Author of the article,
    #     "articleUrl": (string) URL to the article,
    #     "summary": (string) Preview summary of the article,
    #     "appId": (string)* App ID of the calling application,
    #     "appName": (string) App name of the calling application,
    #     "appIconUrl": (string) App icon URL of the calling application
    # }
    def share_room(self, stream_id, content):
        """
        Share third-party content, such as a news article, into the specified stream.
        The stream can be a chat room, an IM, or an MIM.
        content: An object containing the fields required for defining an article

        content = {
            "articleId": (string) A unique ID for this article, not used by any other article,
            "title": (string) The title of the article,
            "subTitle": (string) The subtitle of the article
            "description": (string) The description of the article,
            "message" : (string) The message text that can be sent along with the shared article,
            "publisher": (string)* Publisher of the article,
            "publishDate": (string) Article publish date in unix timestamp,
            "thumbnailUrl": (string) URL to the thumbnail image,
            "author": (string)* Author of the article,
            "articleUrl": (string) URL to the article,
            "summary": (string) Preview summary of the article,
            "appId": (string)* App ID of the calling application,
            "appName": (string) App name of the calling application,
            "appIconUrl": (string) App icon URL of the calling application
        }
        """
        logging.debug('StreamClient/share_room()')
        url = '/agent/v3/stream/{0}/share'.format(stream_id)
        data = {
            "type": "com.symphony.sharing.article",
            "content": content
        }
        return self.bot_client.execute_rest_call('POST', url, json=data)

    def remove_member_from_room(self, stream_id, user_id):
        """
        Removes an existing member from an existing room
        """
        logging.debug('StreamClient/remove_member_from_room()')
        url = '/pod/v1/room/{0}/membership/remove'.format(stream_id)
        data = {'id': user_id}
        return self.bot_client.execute_rest_call('POST', url, json=data)

    def promote_user_to_owner(self, stream_id, user_id):
        """
        Promotes user to owner of the chat room.
        """
        logging.debug('StreamClient/promote_user_to_owner()')
        url = '/pod/v1/room/{0}/membership/promoteOwner'.format(stream_id)
        data = {'id': user_id}
        return self.bot_client.execute_rest_call('POST', url, json=data)

    def demote_user_from_owner(self, stream_id, user_id):
        """
        Demotes room owner to a participant in the chat room.
        """
        logging.debug('StreamClient/demote_user_from_owner()')
        url = '/pod/v1/room/{0}/membership/demoteOwner'.format(stream_id)
        data = {'id': user_id}
        return self.bot_client.execute_rest_call('POST', url, json=data)

    # Available kwargs:
    # labels: A list of room keywords whose values will be queried.
    # active: If true, restricts the search to active rooms. If false, restricts the search to inactive rooms. If not specified, includes both active and inactive rooms.
    # private: If true, includes public and private rooms in the search results. If false or unspecified only public rooms are returned.
    # creator: If provided, restricts the search to rooms created by the specified user.
    # owner: If provided, restricts the search to rooms owned by the specified user.
    # member: If provided, restricts the search to rooms where the specified user is a member.
    # sortOrder: Sort algorithm to be used. Supports two values: BASIC (legacy algorithm) and RELEVANCE (enhanced algorithm).
    def search_rooms(self, query, skip=0, limit=50, **kwargs):
        """
        Search for rooms, querying name, description, and specified keywords.

        search_rooms('query', labels = ['hack', 'test', 'room'], active = True)
        """
        logging.debug('StreamClient/search_rooms()')
        url = '/pod/v3/room/search'
        params = {
            'skip': skip,
            'limit':limit
        }
        data = {
            'query': query
        }
        data.update(kwargs)
        return self.bot_client.execute_rest_call('POST', url, params=params, json=data)

    def get_user_streams(self, skip=0, limit=50, stream_types = 'ALL', include_inactive = True):
        """
        Returns a list of all the streams of which the requesting user is a member,
        sorted by creation date (ascending - oldest to newest).
        """
        logging.debug('StreamClient/get_user_streams()')
        url = '/pod/v1/streams/list'
        if (stream_types == 'ALL'):
            stream_types = [
                    {"type": "IM"},
                    {"type": "MIM"},
                    {"type": "ROOM"},
                    {"type": "POST"}
                ]
        data = {
            'streamTypes': stream_types,
            'includeInactiveStreams': include_inactive
        }
        params = {
            'skip': skip,
            'limit': limit
        }
        return self.bot_client.execute_rest_call('POST', url, json=data, params=params)

    def stream_info_v2(self, stream_id):
        """
        Returns information about a particular stream.
        """
        logging.debug('StreamClient/stream_info_v2()')
        url = '/pod/v2/streams/{0}/info'.format(stream_id)
        return self.bot_client.execute_rest_call('GET', url)


    def list_streams_enterprise(self, skip=0, limit=50, **kwargs):
        """
        Returns a list of all the streams (IMs, MIMs, and chatrooms) for the calling
        user's company, sorted by creation date (ascending – oldest to newest).
        Filtering parameters can be used to narrow the list of streams that are returned.

        Required Permissions: This endpoint may only be called by Service User accounts with the User Provisioning role.

        Available kwargs:

        streamTypes: A list of stream types that will be returned (IM, MIM, ROOM). If not specified, streams of all types are returned.
        scope: The scope of the stream: INTERNAL (restricted to members of the calling user's company) or EXTERNAL (including members of the calling user's company, as well as another firm). If not specified, returns streams of either scope.
        origin: The origin of the room: INTERNAL (created by a user of the calling user's company) or EXTERNAL (created by a user of another company). If not specified, returns streams of either origin. Only applies to chatrooms with External scope.
        privacy: The privacy setting of the room: PRIVATE (members must be added) OR PUBLIC (anyone can join). If not specified, returns both private and public rooms. Only applies to chatrooms with internal scope.
        status: The status of the room: ACTIVE or INACTIVE. If not specified, both active and inactive streams are returned.
        startDate: Restricts result set to rooms that have been modified since this date (an Epoch timestamp specified in milliseconds). When specified along with endDate, enables the developer to specify rooms modified within a given time range.
        endDate: Restricts result set to rooms that have been modified prior to this date (an Epoch timestamp specified in milliseconds). When specified along with startDate, enables the developer to specify rooms modified within a given time range.
        """
        logging.debug('StreamClient/list_streams_enterprise()')
        url = '/pod/v1/admin/streams/list'
        params = {
            'skip': skip,
            'limit': limit
        }
        return self.bot_client.execute_rest_call('POST', url, params=params, json=kwargs)

    def list_streams_enterprise_v2(self, skip=0, limit=50, **kwargs):
        """
        Returns a list of all the streams (IMs, MIMs, and chatrooms) for the calling
        user's company, sorted by creation date (ascending – oldest to newest).
        Filtering parameters can be used to narrow the list of streams that are returned.

        Required Permissions: This endpoint may only be called by Service User accounts with the User Provisioning role.

        Available kwargs:

        streamTypes: A list of stream types that will be returned (IM, MIM, ROOM). If not specified, streams of all types are returned.
        scope: The scope of the stream: INTERNAL (restricted to members of the calling user's company) or EXTERNAL (including members of the calling user's company, as well as another firm). If not specified, returns streams of either scope.
        origin: The origin of the room: INTERNAL (created by a user of the calling user's company) or EXTERNAL (created by a user of another company). If not specified, returns streams of either origin. Only applies to chatrooms with External scope.
        privacy: The privacy setting of the room: PRIVATE (members must be added) OR PUBLIC (anyone can join). If not specified, returns both private and public rooms. Only applies to chatrooms with internal scope.
        status: The status of the room: ACTIVE or INACTIVE. If not specified, both active and inactive streams are returned.
        startDate: Restricts result set to rooms that have been modified since this date (an Epoch timestamp specified in milliseconds). When specified along with endDate, enables the developer to specify rooms modified within a given time range.
        endDate: Restricts result set to rooms that have been modified prior to this date (an Epoch timestamp specified in milliseconds). When specified along with startDate, enables the developer to specify rooms modified within a given time range.
        """
        logging.debug('StreamClient/list_streams_enterprise_v2()')
        url = '/pod/v2/admin/streams/list'
        params = {
            'skip': skip,
            'limit': limit
        }
        return self.bot_client.execute_rest_call('POST', url, params=params, json=kwargs)

    def get_stream_members(self, stream_id, skip=0, limit=100):
        """
        Returns a list of all the current members of a stream (IM, MIM, or chatroom
        Required Permissions:

        To get the stream membership of any stream in your enterprise, you should
        call this endpoint with a Service User account with the User Provisioning role.
        The Service User does not need to be a member of the stream.
        See Permissions for a list of roles and associated privileges.
        """
        logging.debug('StreamClient/get_stream_members()')
        url = '/pod/v1/admin/stream/{0}/membership/list'.format(stream_id)
        params = {
            'skip': skip,
            'limit': limit
        }
        return self.bot_client.execute_rest_call('GET', url, params=params)



import json
import logging
from json.decoder import JSONDecodeError

import aiohttp
import requests
import ssl

from .admin_client import AdminClient
from .api_client import APIClient
from .connections_client import ConnectionsClient
from .datafeed_client import DataFeedClient
from .health_check_client import HealthCheckClient
from .message_client import MessageClient
from .signals_client import SignalsClient
from .stream_client import StreamClient
from .user_client import UserClient
from ..datafeed_event_service import AsyncDataFeedEventService, DataFeedEventService
from ..exceptions.UnauthorizedException import UnauthorizedException

# SymBotClient class is the Client class that has access to all of the other
# client classes upon initialization, SymBotClient class gets an instance of
# each client along with access to all of its methods.
# class contains series of getters for each client
# class also contains config and auth class as a way to pass this info around
# to each client as well class is seen as orchestrator or interface for all
# REST API calls

# Saving this has as a constant because it's easily typoed and used everywhere
_TRUSTSTORE_PATH = "truststorePath"


class SymBotClient(APIClient):

    def __init__(self, auth, config):
        self.auth = auth
        self.config = config
        self.agentConfig = config
        self.datafeed_event_service = None
        self.async_datafeed_event_service = None
        self.datafeed_client = None
        self.msg_client = None
        self.stream_client = None
        self.user_client = None
        self.connections_client = None
        self.signals_client = None
        self.admin_client = None
        self.api_client = None
        self.pod_session = None
        self.async_pod_session = None
        self.agent_session = None
        self.async_agent_session = None
        self.bot_user_info = None
        self.health_check_client = None
        self.async_ssl_context = None


    def get_datafeed_event_service(self, *args, **kwargs):
        if self.datafeed_event_service is None:
            self.datafeed_event_service = DataFeedEventService(self, *args, **kwargs)
        return self.datafeed_event_service

    def get_async_datafeed_event_service(self, *args, **kwargs):
        if self.async_datafeed_event_service is None:
            self.async_datafeed_event_service = AsyncDataFeedEventService(self, *args, **kwargs)
        return self.async_datafeed_event_service

    def get_datafeed_client(self):
        if self.datafeed_client is None:
            self.datafeed_client = DataFeedClient(self)
        return self.datafeed_client

    def get_message_client(self):
        if self.msg_client is None:
            self.msg_client = MessageClient(self)
        return self.msg_client

    def get_admin_client(self):
        if self.admin_client is None:
            self.admin_client = AdminClient(self)
        return self.admin_client

    def get_signals_client(self):
        if self.signals_client is None:
            self.signals_client = SignalsClient(self)
        return self.signals_client

    def get_connections_client(self):
        if self.connections_client is None:
            self.connections_client = ConnectionsClient(self)
        return self.connections_client

    def get_stream_client(self):
        if self.stream_client is None:
            self.stream_client = StreamClient(self)
        return self.stream_client

    def get_user_client(self):
        if self.user_client is None:
            self.user_client = UserClient(self)
        return self.user_client

    def get_api_client(self):
        self.api_client = APIClient(self)

    def get_sym_config(self):
        return self.config

    def get_sym_agent_config(self):
        return self.agentConfig

    def get_sym_auth(self):
        return self.auth

    def get_pod_session(self):
        """This is the method to retrieve the session object for synchronous calls with requests"""
        if self.pod_session is None:
            logging.debug('bot_client/get_pod_session() - creating pod session')
            self.pod_session = requests.Session()
            self.pod_session.headers.update({
                'sessionToken': self.auth.get_session_token(),
                'cache-control': 'no-cache'}
            )
            self.pod_session.proxies.update(self.config.data['podProxyRequestObject'])
            if self.config.data[_TRUSTSTORE_PATH]:
                logging.debug("Setting truststorePath for pod to {}".format(
                    self.config.data[_TRUSTSTORE_PATH]))
                self.pod_session.verify = self.config.data[_TRUSTSTORE_PATH]
        return self.pod_session

    def get_agent_session(self):
        """This is the method to retrieve the session object for synchronous calls with requests"""
        if self.agent_session is None:
            logging.debug('bot_client/get_agent_session() - creating agent session')
            self.agent_session = requests.Session()
            self.agent_session.headers.update({
                'sessionToken': self.auth.get_session_token(),
                'keyManagerToken': self.auth.get_key_manager_token(),
                'cache-control': 'no-cache'}
            )
            self.agent_session.proxies.update(self.config.data['agentProxyRequestObject'])
            if self.config.data[_TRUSTSTORE_PATH]:
                logging.debug("Setting truststorePath for agent to {}".format(
                    self.config.data[_TRUSTSTORE_PATH])
                )
                self.agent_session.verify=self.config.data[_TRUSTSTORE_PATH]

        return self.agent_session

    def execute_rest_call(self, method, path, **kwargs):
        results = None
        session = None
        if path.startswith("/agent/"):
            url = self.config.data["agentUrl"] + path
            session = self.get_agent_session()
        elif path.startswith("/pod/"):
            url = self.config.data["podUrl"] + path
            session = self.get_pod_session()
        else:
            url = path
            session = self.get_agent_session()

        try:
            response = session.request(method, url, **kwargs)
        except requests.exceptions.ConnectionError as err:
            logging.debug(err)
            logging.debug(type(err))
            logging.debug('ensure pod/agent subdomains are correct')
            raise
        if response.status_code == 204:
            results = []
        elif response.status_code == 200 or response.status_code == 201:
            try:
                results = json.loads(response.text)
            except JSONDecodeError:
                results = response.text
        else:
            # Try to get the json to be used to handle the error message
            error_json = None
            text = None
            try:
                error_json = response.json()
            except JSONDecodeError:
                try:
                    text = response.text
                except Exception:
                    text = None
            try:
                super().handle_error(response, self, error_json, text)
            except UnauthorizedException:
                logging.debug('caught UnauthorizedException - try execute_rest_call() again')
                self.execute_rest_call(method, path, **kwargs)
        return results


    def get_async_pod_session(self):
        """This is the method to retrieve the session object for asynchronous calls with aiohttp"""
        if self.async_pod_session is None:
            logging.debug('bot_client/get_pod_session() - creating async pod session')
            self.async_pod_session = aiohttp.ClientSession(
            headers = {
                'sessionToken': self.auth.get_session_token(),
                'cache-control': 'no-cache'}
                )
            # For aiohttp proxies and truststore are handled when the request is made
        return self.async_pod_session

    def get_async_agent_session(self):
        """This is the method to retrieve the session object for asynchronous calls with aiohttp"""
        if self.async_agent_session is None:
            logging.debug('bot_client/get_agent_session() - creating async agent session')
            self.async_agent_session = aiohttp.ClientSession(
            headers = {
                'sessionToken': self.auth.get_session_token(),
                'keyManagerToken': self.auth.get_key_manager_token(),
                'cache-control': 'no-cache'}
            )
            # For aiohttp proxies and truststore are handled when the request is made
        return self.async_agent_session

    # Known issue on this function when using a proxy due to an outstanding issue with aiohttp
    # To workaround this please check README.md
    async def execute_rest_call_async(self, method, path, **kwargs):
        """This is the asynchronous method to hit the rest api, it should be awaited"""
        results = None
        session = None

        if path.startswith("/agent/"):
            url = self.config.data["agentUrl"] + path
            session = self.get_async_agent_session()
            http_proxy = self.config.data['agentProxyRequestObject'].get("http")

        elif path.startswith("/pod/"):
            url = self.config.data["podUrl"] + path
            session = self.get_async_pod_session()
            http_proxy = self.config.data['podProxyRequestObject'].get("http")
        else:
            # TODO: Confirm whether this should just throw
            # Not sure what the best course of action is here, taking pod values
            url = path
            session = self.get_async_pod_session()
            http_proxy = self.config.data['podProxyRequestObject'].get("http")

        # This is to handle the files keyword
        files = kwargs.pop("files", None)

        # The below attempts to handle a files kwarg in the same way that Requests handles it
        if files is not None:
            with aiohttp.MultipartWriter("form-data") as mpwriter:
                for (key, value) in files.items():
                    part = mpwriter.append(value)
                    part.set_content_disposition("form-data", name=key)
            if kwargs.get("data"):
                # This isn't fatal, it's just not yet clear how to handle this in aiohttp
                # This link is the best resource I've seen explaining the issue
                # https://github.com/aio-libs/aiohttp/issues/3571
                raise RuntimeError("Not expecting to find data and files")
            else:
                data = mpwriter
        elif kwargs.get("data") is not None:
            data = kwargs.pop("data")
        else:
            data = None


        try:
            response = await session.request(method, url, proxy=http_proxy, ssl=self.get_async_ssl_context(), data=data, **kwargs)
        except aiohttp.ClientConnectionError as err:
            logging.debug(err)
            logging.debug(type(err))
            logging.debug('ensure pod/agent subdomains are correct')
            raise

        if response.status == 204:
            results = []
        elif response.status == 200:
            text = await response.text()

            try:
                results = json.loads(text)
            except JSONDecodeError:
                results = text
        else:
            # Try to get the json to be used to handle the error message
            error_json = None
            text = None
            try:
                error_json = await response.json()
            except JSONDecodeError:
                try:
                    text = await response.text()
                except Exception:
                    text = None
            try:
                super().handle_error(response, self, error_json, text)
            except UnauthorizedException:
                await self.execute_rest_call_async(method, path, **kwargs)
        return results

    def reauth_client(self):
        self.auth.authenticate()
        if self.pod_session:
            logging.debug('bot_client/reauth_client() - pod session exists')
            self.pod_session.headers.update({
                'sessionToken': self.auth.get_session_token()}
            )
        if self.agent_session:
            logging.debug('bot_client/reauth_client() - agent session exists')
            self.agent_session.headers.update({
                'sessionToken' : self.auth.get_session_token(),
                'keyManagerToken': self.auth.get_key_manager_token()}
            )

        if self.async_pod_session:
            logging.debug('bot_client/reauth_client() - async pod session exists')
            self.async_pod_session = aiohttp.ClientSession(
            headers = {
                'sessionToken': self.auth.get_session_token()}
            )

        if self.async_agent_session:
            logging.debug('bot_client/reauth_client() - async agent session exists')
            self.async_agent_session = aiohttp.ClientSession(
            headers = {
                'sessionToken': self.auth.get_session_token(),
                'keyManagerToken': self.auth.get_key_manager_token()}
            )

    def get_bot_user_info(self):
        if self.bot_user_info is None:
            self.bot_user_info = self.get_user_client().get_session_user()
        return self.bot_user_info

    def get_health_check_client(self):
        if self.health_check_client is None:
            self.health_check_client = HealthCheckClient(self)
        return self.health_check_client

    #Required for aiohttp to use truststore
    def get_async_ssl_context(self):
        if self.async_ssl_context is None:
            if self.config.data[_TRUSTSTORE_PATH]:
                logging.debug("Setting truststorePath for async calls to {}".format(self.config.data[_TRUSTSTORE_PATH]))
                self.async_ssl_context = ssl.create_default_context(cafile=self.config.data[_TRUSTSTORE_PATH])
            else:
                self.async_ssl_context = ssl.create_default_context()
        return self.async_ssl_context

    async def close_async_sessions(self):
        """Close the open aiohttp.ClientSession objects

        In general this shouldn't be necessary as the interpreter should tear down
        any Python objects, and subsequently the OS should terminate any other
        resources (in this case sockets). However due to bugs in the Python standard
        library certain situations trigger a race condition between asyncio and the
        logging library that clobbers the stacktrace and raises a NameError. This
        just ensures that doesn't happen and gives a cleaner output from the SDK.

        For most usecases this method can be safely omitted.
        """
        logging.debug("Manually closing sessions")
        if self.async_pod_session:
            await self.async_pod_session.close()
            self.async_pod_session = None

        if self.async_agent_session:
            await self.async_agent_session.close()
            self.async_agent_session = None



import logging
from .api_client import APIClient


class DataFeedClientV1(APIClient):
    def __init__(self, bot_client):
        self.bot_client = bot_client

    # raw api call to create_datafeed --> returns datafeed_id
    def create_datafeed(self):

        url = '/agent/v4/datafeed/create'
        response = self.bot_client.execute_rest_call("POST", url)
        datafeed_id = response.get('id')
        logging.debug('DataFeedClientV1/create_datafeed() --> {}'.format(datafeed_id))
        return datafeed_id


    def read_datafeed(self, datafeed_id, *ackId):
        logging.debug('DataFeedClientV1/read_datafeed()')
        url = '/agent/v4/datafeed/{0}/read'.format(datafeed_id)
        datafeed_read = self.bot_client.execute_rest_call("GET", url)
        return datafeed_read

    def list_datafeed_id(self):
        logging.debug('DataFeedClientV1/list_datafeed_id()')
        raise TypeError("This function is not supported for the DF V1 client.")

    def delete_datafeed(self, datafeed_id):
        logging.debug('DataFeedClientV1/delete_datafeed()')
        raise TypeError("This function is not supported for the DF V1 client.")

    def get_ack_id(self):
        logging.debug('DataFeedClientV1/get_ack_id()')
        raise TypeError("This function is not supported for the DF V1 client.")


    async def read_datafeed_async(self, datafeed_id):

        logging.debug('DataFeedClientV1/read_datafeed_async()')
        url = '/agent/v4/datafeed/{0}/read'.format(datafeed_id)
        datafeed_read = await self.bot_client.execute_rest_call_async("GET", url)

        return datafeed_read




from abc import ABC, abstractmethod

# Abstract class for IM listener
# class is just an interface of functions to handle the IM events received
# from DataFeed
# (see Real Time Events in REST API documentation for more details)
# the developer will handle actual event logic in your implementation of
# this abstract class


class IMListener(ABC):

    @abstractmethod
    def on_im_message(self, message):
        """
        Do Something
        """

    @abstractmethod
    def on_im_created(self, stream):
        """
        Do Something
        """



import logging

import aiohttp
import requests
from requests_toolbelt.multipart.encoder import MultipartEncoder

from ..exceptions.APIClientErrorException import APIClientErrorException
from ..exceptions.DatafeedExpiredException import DatafeedExpiredException
from ..exceptions.ForbiddenException import ForbiddenException
from ..exceptions.ServerErrorException import ServerErrorException
from ..exceptions.UnauthorizedException import UnauthorizedException


# error handling class --> take status code and raise appropriate exceptions
# this class acts as a parent class to each of the other client class.
# each child class extends error handling functionality


class APIClient:

    def __init__(self, bot_client):
        self.bot_client = bot_client

    def make_mulitpart_form(self, fields, aio=False):
        """Create a multipart form to be used across the Symphony API, that works for both requests
        and the asynchronous aiohttp. Requests basically uses requests-toolbelt, but it's a little
        bit more involved for aiohttp. The output of this is expected to be passed to either
        execute_rest_request or execute_rest_request_async depending whether aio was true"""

        if aio:
            # This appears to be the canonical way to use aiohttp to pass mulipart data into the API
            # in the same way that MultipartEncoder does for Requests.
            # aiohttp.FormData does appear to work because of the way the Symphony API demands a boundary
            # in the header. aiohttp.MultipartWriter.append_form doesn't appear to work because it
            # encodes as a application/x-www-form-urlencoded that Symphony doesn't appear to like for
            # attachments
            with aiohttp.MultipartWriter("form-data") as data:
                for k, v in fields.items():
                    if len(v) == 1:
                        part = data.append(v)
                        part.set_content_disposition("form-data", name=k)
                    if len(v) == 3:
                        filename, file_object, content_type = v
                        part = data.append(file_object, {'Content-Type': content_type})
                        part.set_content_disposition('form-data', name=k, filename=filename)

            headers = {
                'Content-Type': 'multipart/form-data; boundary=' + data.boundary
            }

        else:
            data = MultipartEncoder(
                fields=fields
            )
            headers = {
                'Content-Type': data.content_type
            }

        return {"data": data, "headers": headers}

    def handle_error(self, response, bot_client, error_json=None, text=None):
        logging.debug('api_client/handle_error() function started')
        _error_field = "message"
        if isinstance(response, requests.Response):
            status = response.status_code
        else:
            # The assumption is that it's an aiohttp response from an async request
            status = response.status

        try:
            if error_json is not None:
                try:
                    err_message = error_json[_error_field]
                except KeyError:
                    if text is not None:
                        err_message = text
                    else:
                        err_message = ""
            elif text is not None:
                err_message = text
            else:
                err_message = ""

        except Exception:
            logging.error("Unable to parse error message: {}".format(text))
            err_message = ""

        if status == 400 and 'Could not find a datafeed with the' in err_message:
            logging.debug('datafeed expired, start_datafeed()')
            raise DatafeedExpiredException()

        elif status == 401:
            logging.debug('api_client()/handling 401 error')
            bot_client.reauth_client()
            logging.debug('api_client()/successfully reauthenticated')
            raise UnauthorizedException(
                'User, unauthorized, refreshing tokens: {}'
                    .format(status))
        elif status == 403:
            raise ForbiddenException(
                'Forbidden: Caller lacks necessary entitlement: {}'
                    .format(status))
        elif status == 405:
            logging.debug('Handling 405 error')
            raise ForbiddenException(
                'Method Not Allowed: The method received in the request-line is known by the origin server but not supported by the target resource: {}'
                    .format(status))

        # Response dict is a bit of an information overload, could consider trimming it
        elif 400 <= status < 500:

            raise APIClientErrorException('Client Error Occurred: {}. Response contents: {}'
                                          .format(err_message, response.__dict__))

        elif status >= 500:
            raise ServerErrorException(
                'Server Error Exception: {}, {}'
                    .format(status, err_message))
        else:
            # This shouldn't happen
            raise RuntimeError("Unhandled error: {}".format(response.__dict__))



import logging

from .api_client import APIClient


# child class of APIClient --> Extends error handling functionality
# PresenceClient class contains a series of functions corresponding to all
# pod admin endpoints on the REST API.
class PresenceClient(APIClient):

    def __init__(self, bot_client):
        self.bot_client = bot_client

    def get_presence(self):
        """Returns the online status of the calling user."""
        logging.debug('PresenceClient/get_presence()')
        url = '/pod/v2/user/presence/'

        return self.bot_client.execute_rest_call('GET', url)

    def get_all_presence(self, last_user_id, limit):
        """
        Returns the presence of all users in a pod

        lastUserId - Last user ID retrieved, used for paging. If provided, results skip users with IDs less than this parameter.

        limit - Maximum number of records to return. The maximum supported value is 5000.

        All non-inactive users are returned and some inactive users may be included. Any omitted user is inactive.

        Only users with the User Provisioning role can call this endpoint
        """
        logging.debug('PresenceClient/get_all_presence()')
        url = '/pod/v2/users/presence'
        params = {
                'lastUserId': last_user_id,
                'limit': limit
                }
        return self.bot_client.execute_rest_call('GET', url, params=params)

    def get_user_status(self, user_id, local=True):
        """
        Returns the online status of the specified user.


        local
        true: Perform a local query and set the presence to OFFLINE for users who are not local to the calling user’s pod.
        false or absent: Query the presence of all local and external users who are connected to the calling user. See also Query Presence of External Users.
        
        The timestamp in the response is in UTC format and contains the time when the presence was set using Set Presence. For users who are offline, the timestamp contains the current time when the Get User Presence endpoint was invoked.

        To get the presence of external users, you must first register interest about their presence.

        When calling this as an OBO-enabled endpoint, use the OBO User Authenticate token for sessionToken.

    The available online status values (presence categories) for users are listed on the Get Presence page.

        Query Presence of External Users
    To query the presence of external users, you must first perform an additional step:
        1. Call the Register Interest in External User Presence endpoint to register interest in those users.
        2. Call this endpoint to query the presence of each user.

        To query the presence of internal users, you do not need to register interest and can call this endpoint directly.
        """
        logging.debug('PresenceClient/get_user_status()')
        url = '/pod/v3/user/{0}/presence'.format(user_id)
        params = {'local': local}
        return self.bot_client.execute_rest_call("GET", url, params=params)

    def external_presence_interest(self, array_user_ids):
        """
        To get the presence state of external users, you must first register interest in those users using this endpoint.

        Any user can see the presence of other users of the same company. For users of a different company, the two users must be connected to see presence.

        To query the presence of external users:
        1. Call this endpoint to register interest in the desired users.
        2. Call the Get User Presence endpoint to query the presence of each user.
        To keep the registration active, call this endpoint every hour.

        To query the presence of internal users, you do not need to register interest.
        Rate Limit
Getting an external user’s presence is limited to one call every 5 minutes.

        Roles and Privileges
        Calling this endpoint requires the ADMIN_PRESENCE_UPDATE privilege.
        See Permissions for a list of roles and associated privileges.
        """
        logging.debug('PresenceClient/external_presence_interest()')
        url = '/pod/v1/user/presence/register'
        data = {'userIds': array_user_ids}
        return self.bot_client.execute_rest_call("POST", url, json=data)

    def set_presence(self, category):
        """
        Sets the online status of the calling user.

        category*
        The new presence state for the user. Possible values are AVAILABLE, BUSY, AWAY, ON_THE_PHONE, BE_RIGHT_BACK, IN_A_MEETING, OUT_OF_OFFICE, OFF_WORK.
        
        Calling this endpoint requires the ADMIN_PRESENCE_UPDATE privilege.
        See Permissions for a list of roles and associated privileges.
        """
        logging.debug('PresenceClient/set_presence()')
        url = '/pod/v2/user/presence'
        data = {'category': category}
        return self.bot_client.execute_rest_call("POST", url, json=data)

    def create_presence_feed(self):
        """
        Creates a new stream capturing online status changes ("presence feed") for the company (pod) and returns the ID of the new feed. 
        The feed will return the presence of users whose presence status has changed since it was last read.
        """
        logging.debug('PresenceClient/create_presence_feed()')
        url = '/pod/v1/presence/feed/create'
        return self.bot_client.execute_rest_call("POST", url)

    def read_presence_feed(self, feed_id):
        """
        Reads the specified presence feed that was created using the Create Presence feed endpoint. 
        The feed returned includes the user presence statuses that have changed since they were last read.
        """
        logging.debug('PresenceClient/read_presence_feed()')
        url = '/pod/v1/presence/feed/{0}/read'.format(feed_id)
        return self.bot_client.execute_rest_call("GET", url)


    def delete_presence_feed(self, feed_id):
        """
        Deletes a presence status feed. 
        This endpoint returns the ID of the deleted feed if the deletion is successful.
        """
        logging.debug('PresenceClient/delete_presence_feed()')
        url = '/pod/v1/presence/feed/{0}/delete'.format(feed_id)
        return self.bot_client.execute_rest_call("GET", url)

    def set_user_presence(self, user_id, category):
        """
        Sets the presence state of a another user.

        To set the presence of another user, this endpoint requires to be called by a service user with the "User Provisioning" Entitlement.
        """
        logging.debug('PresenceClient/set_user_presence()')
        url = '/pod/v3/user/presence'
        data = {
                'category': category,
                'userId': user_id
                }
        return self.bot_client.execute_rest_call("POST", url, json=data)



import logging
from .api_client import APIClient
import logging

from .api_client import APIClient


# logging.basicConfig(filename='logs/example.log', format='%(asctime)s - %(
# name)s - %(levelname)s - %(message)s', filemode='w', level=logging.DEBUG)
# logging.getLogger("urllib3").setLevel(logging.WARNING)

# child class of APIClient --> Extends error handling functionality
# UserClient class contains a series of functions corresponding to all
# messaging
# endpoints on the REST API.


class UserClient(APIClient):

    def __init__(self, bot_client):
        self.bot_client = bot_client

    def get_user_from_user_name(self, user_name):
        logging.debug('UserClient/get_user_from_user_name()')
        url = '/pod/v2/user'
        params = {'username': user_name}
        return self.bot_client.execute_rest_call('GET', url, params=params)

    def get_user_from_email(self, email, local=False):
        logging.debug('UserClient/get_user_from_email()')
        url = '/pod/v2/user'
        params = {'email': email, 'local':local}
        return self.bot_client.execute_rest_call('GET', url, params=params)

    def get_user_from_id(self, user_id, local=False):
        logging.debug('UserClient/get_user_from_id()')
        url = '/pod/v2/user'
        params = {'uid': user_id, 'local':local}
        return self.bot_client.execute_rest_call('GET', url, params=params)

    def get_users_from_id_list(self, user_id_list, local=False):
        logging.debug('UserClient/get_users_from_id_list()')
        url = '/pod/v3/users'
        users_array = ','.join(map(str, user_id_list))
        params = {'uid': users_array, 'local': local}
        return self.bot_client.execute_rest_call('GET', url, params=params)

    def get_users_from_email_list(self, email_list, local=False):
        logging.debug('UserClient/get_users_from_email_list()')
        url = '/pod/v3/users'
        usersArray = ','.join(map(str, email_list))
        params = {'email': usersArray, 'local': local}
        return self.bot_client.execute_rest_call('GET', url, params=params)

    def search_users(self, query, local=False, skip=0, limit=50, filters={}):
        logging.debug('UserClient/search_users()')
        url = '/pod/v1/user/search'
        params = {'local': local, 'skip': skip, 'limit': limit}
        data = {'query':query, 'filters': filters}
        return self.bot_client.execute_rest_call('POST', url, params=params, json=data)

    def get_session_user(self):
        logging.debug('UserClient/get_session_user()')
        url = '/pod/v2/sessioninfo'
        return self.bot_client.execute_rest_call('GET', url)



import json
import logging
import time
import requests
from .auth_endpoint_constants import auth_endpoint_constants
from ..clients.api_client import APIClient
from requests_pkcs12 import Pkcs12Adapter
from ..exceptions.UnauthorizedException import UnauthorizedException
from ..exceptions.MaxRetryException import MaxRetryException


class Auth(APIClient):
    """Class for certificate authentication"""

    def __init__(self, config):
        """
        Set up proxy information if configuration contains proxyURL
        :param config: Object contains all certificate configurations
        """
        self.config = config
        self.agentConfig = config
        self.last_auth_time = 0
        self.auth_retries = 0
        self.session_token = None
        self.key_manager_token = None
        self.auth_session = requests.Session()
        self.key_manager_auth_session = requests.Session()

        # proxy infomation set in config loader, set to empty object if there is no proxy set in config.json
        self.auth_session.proxies.update(self.config.data['podProxyRequestObject'])
        self.key_manager_auth_session.proxies.update(self.config.data['keyManagerProxyRequestObject'])

        if self.config.data['truststorePath']:
            logging.debug('truststore being added to requests library')
            self.auth_session.verify = self.config.data['truststorePath']
            self.key_manager_auth_session.verify = self.config.data['truststorePath']

        self.auth_session.mount(
            self.config.data['sessionAuthUrl'],
            Pkcs12Adapter(
                pkcs12_filename=self.config.data['p.12'],
                pkcs12_password=self.config.data['botCertPassword']
            ))
        self.auth_session.mount(
            self.config.data['keyAuthUrl'],
            Pkcs12Adapter(
                pkcs12_filename=self.config.data['p.12'],
                pkcs12_password=self.config.data['botCertPassword']
            ))

        self.key_manager_auth_session.mount(
            self.config.data['sessionAuthUrl'],
            Pkcs12Adapter(
                pkcs12_filename=self.config.data['p.12'],
                pkcs12_password=self.config.data['botCertPassword']
            ))
        self.key_manager_auth_session.mount(
            self.config.data['keyAuthUrl'],
            Pkcs12Adapter(
                pkcs12_filename=self.config.data['p.12'],
                pkcs12_password=self.config.data['botCertPassword']
            ))

    def get_session_token(self):
        """Return the session token"""
        return self.session_token

    def get_key_manager_token(self):
        """Return the key manager token"""
        return self.key_manager_token

    def authenticate(self):
        """
        Get the session and key manager token
        """
        logging.debug('Auth/authenticate()')
        try:
            if (self.last_auth_time == 0) or \
                    (int(round(time.time() * 1000) - self.last_auth_time >= auth_endpoint_constants['WAIT_TIME'])):
                logging.debug('Auth/authenticate() --> needed to authenticate')

                self.last_auth_time = int(round(time.time() * 1000))
                self.session_authenticate()
                self.key_manager_authenticate()

            else:
                logging.debug('Retry authentication in 30 seconds.')
                time.sleep(auth_endpoint_constants['TIMEOUT'])
                self.authenticate()
        except:
            raise MaxRetryException('max auth retry limit')

    # Retrieve session token by calling the session token API
    # Certificates are passed in cert parameter
    def session_authenticate(self):
        """
        Get the session token by calling API, certificate data is
        passed in through Request Session object
        """
        logging.debug('Auth/get_session_token()')

        url = self.config.data['sessionAuthUrl'] + '/sessionauth/v1/authenticate'
        response = self.auth_session.post(url)

        if response.status_code != 200:
            self.auth_retries += 1
            if self.auth_retries > auth_endpoint_constants['MAX_AUTH_RETRY']:
                logging.debug('more than 5 times tried')
                raise UnauthorizedException('max auth retry limit: {}'.format(response.__dict__))
            else:
                logging.debug('Auth/get_session_token() function failed: {}'.format(
                    response.status_code)
                )
                time.sleep(auth_endpoint_constants['TIMEOUT'])
                self.session_authenticate()
        else:
            data = json.loads(response.text)
            logging.debug('Auth/session token success')
            self.session_token = data['token']
            self.auth_retries = 0

    def key_manager_authenticate(self):
        """
        Get the key manager token by calling API, certificate data is
        passed in through Request Session object
        """
        logging.debug('Auth/get_keyauth()')
        url = self.config.data['keyAuthUrl'] + '/keyauth/v1/authenticate'
        response = self.key_manager_auth_session.post(url)

        if response.status_code != 200:
            self.auth_retries += 1
            if self.auth_retries > auth_endpoint_constants['MAX_AUTH_RETRY']:
                raise UnauthorizedException('max auth retry limit: {}'.format(response.__dict__))
            else:
                logging.debug('Auth/get_key_manager_authenticate() function failed: {}'.format(
                    response.status_code)
                )
                time.sleep(auth_endpoint_constants['TIMEOUT'])
                self.key_manager_authenticate()
        else:
            data = json.loads(response.text)
            logging.debug('Auth/key manager token success')
            self.key_manager_token = data['token']
            self.auth_retries = 0



import json
import logging
import os

from sym_api_client_python.clients.constants.DatafeedVersion import DatafeedVersion


class SymConfig:
    # initialize object by passing in path to config file
    # store configuration data in variable data
    def __init__(self, path_to_config, relative_to=None):
        """If relative_to is supplied, all relative paths will be recomputed to be relative to the file or directory
        supplied. This allows starting the bots from other directories than where the main is. If a directory is
        given in this field, it should end with a slash.

        In testing one may want to set relative_to to be the path to the config json, so all references are relative
        to that. An application may wish to set this to __file__ in its main, meaning a configuration file from anywhere
        could be used.

        """
        self.path_to_config = path_to_config
        self.relative_to = os.path.dirname(relative_to) if relative_to is not None else os.curdir
        self.data = {}

    def _fix_relative_path(self, json_data, path_key, filename_key=None, warn_if_absent=True):
        """Given a json file, a key for a path and an optional key for a filename, extract the path
        and potentially name, resolve and join them. If warn_if_absent, issue a warning if the file
        or path does not exist at that location"""

        path = json_data[path_key]

        # Blank values are used to ignore entry. They should probably be None instead of blank but to maintain
        # backwards compatibility the function just returns "". If it continued "" would get resolved to "."
        if path == "":
            return ""

        if filename_key is not None:
            filename = json_data[filename_key]
            path = os.path.join(path, filename)
        result = os.path.normpath(os.path.join(self.relative_to, path))

        if warn_if_absent and (not os.path.exists(result)):
            parts = [p for p in [path_key, filename_key] if p is not None]
            logging.warning(
                "{} specified in config, but resolved path {} does not exist"
                    .format(", ".join(parts), result))
        return result

    def load_config(self):
        logging.debug("Loading config from: {}".format(os.path.realpath(self.path_to_config)))
        with open(self.path_to_config, "r") as read_file:
            data = json.load(read_file)
            self.data = data

            self.data['sessionAuthUrl'] = self._build_url(data.get('sessionAuthHost'), data.get('sessionAuthPort'),
                                                          data.get('sessionAuthContextPath'))
            self.data['keyAuthUrl'] = self._build_url(data.get('keyAuthHost'), data.get('keyAuthPort'),
                                                      data.get('keyAuthContextPath'))
            self.data['podUrl'] = self._build_url(data.get('podHost'), data.get('podPort'), data.get('podContextPath'))
            self.data['agentUrl'] = self._build_url(data.get('agentHost'), data.get('agentPort'),
                                                    data.get('agentContextPath'))

            # re-assign url to host to keep backward compatibility
            self.data['sessionAuthHost'] = self.data['sessionAuthUrl']
            self.data['keyAuthHost'] = self.data['keyAuthUrl']
            self.data['podHost'] = self.data['podUrl']
            self.data['agentHost'] = self.data['agentUrl']

            # backwards compatible
            if 'botCertPath' in data:
                self.data['botCertPath'] = self._fix_relative_path(data, 'botCertPath')

            if 'botRSAName' in data:
                self.data['botRSAPath'] = self._fix_relative_path(data, 'botRSAPath', 'botRSAName')

            if 'botPrivateKeyName' in data:
                self.data['botRSAPath'] = self._fix_relative_path(data, 'botPrivateKeyPath', 'botPrivateKeyName')

            if 'botCertName' in data:
                self.data['p.12'] = self._fix_relative_path(data, 'botCertPath', 'botCertName')

            if 'truststorePath' in data:
                self.data['truststorePath'] = self._fix_relative_path(data, 'truststorePath')

            if 'proxyURL' in data:
                self.data['podProxyURL'] = data['proxyURL']
                self.data['podProxyUsername'] = data['proxyUsername'] if 'proxyUsername' in data else ""
                self.data['podProxyPassword'] = data['proxyPassword'] if 'proxyPassword' in data else ""
                self.data['agentProxyURL'] = data['proxyURL']
                self.data['agentProxyUsername'] = data['proxyUsername'] if 'proxyUsername' in data else ""
                self.data['agentProxyPassword'] = data['proxyPassword'] if 'proxyPassword' in data else ""
                self.data['keyManagerProxyURL'] = data['proxyURL']
                self.data['keyManagerProxyUsername'] = data['proxyUsername'] if 'proxyUsername' in data else ""
                self.data['keyManagerProxyPassword'] = data['proxyPassword'] if 'proxyPassword' in data else ""

            if 'podProxyURL' not in data or not data['podProxyURL']:
                self.data['podProxyRequestObject'] = {}
                self.data['podProxyURL'] = ""
            else:
                self.data['podProxyURL'] = data['podProxyURL']

                if 'podProxyUsername' in data and data['podProxyUsername']:
                    self.data['podProxyUsername'] = data['podProxyUsername']
                    self.data['podProxyPassword'] = data['podProxyPassword']
                    pod_proxy_parse = data['podProxyURL'].split('://')
                    pod_proxy_auth = data['podProxyUsername'] + ':' + data['podProxyPassword']
                    pod_proxy_url = pod_proxy_parse[0] + '://' + pod_proxy_auth + '@' + pod_proxy_parse[1]
                    self.data['podProxyRequestObject'] = {
                        'http': pod_proxy_url,
                        'https': pod_proxy_url,
                    }
                else:
                    self.data['podProxyRequestObject'] = {
                        'http': data['podProxyURL'],
                        'https': data['podProxyURL'],
                    }

            if 'agentProxyURL' not in data or not data['agentProxyURL']:
                self.data['agentProxyRequestObject'] = {}
                self.data['agentProxyURL'] = ""
            else:
                self.data['agentProxyURL'] = data['agentProxyURL']

                if 'agentProxyUsername' in data and data['agentProxyUsername']:
                    self.data['agentProxyUsername'] = data['agentProxyUsername']
                    self.data['agentProxyPassword'] = data['agentProxyPassword']
                    agent_proxy_parse = data['agentProxyURL'].split('://')
                    agent_proxy_auth = data['agentProxyUsername'] + ':' + data['agentProxyPassword']
                    agent_proxy_url = agent_proxy_parse[0] + '://' + agent_proxy_auth + '@' + agent_proxy_parse[1]
                    self.data['agentProxyRequestObject'] = {
                        'http': agent_proxy_url,
                        'https': agent_proxy_url,
                    }
                else:
                    self.data['agentProxyRequestObject'] = {
                        'http': data['agentProxyURL'],
                        'https': data['agentProxyURL'],
                    }

            if 'keyManagerProxyURL' not in data or not data['keyManagerProxyURL']:
                self.data['keyManagerProxyRequestObject'] = {}
                self.data['keyManagerProxyURL'] = ""
            else:
                self.data['keyManagerProxyURL'] = data['keyManagerProxyURL']

                if 'keyManagerProxyUsername' in data and data['keyManagerProxyUsername']:
                    self.data['keyManagerProxyUsername'] = data['keyManagerProxyUsername']
                    self.data['keyManagerProxyPassword'] = data['keyManagerProxyPassword']
                    km_proxy_parse = data['keyManagerProxyURL'].split('://')
                    km_proxy_auth = data['keyManagerProxyUsername'] + ':' + data['keyManagerProxyPassword']
                    km_proxy_url = km_proxy_parse[0] + '://' + km_proxy_auth + '@' + km_proxy_parse[1]
                    self.data['keyManagerProxyRequestObject'] = {
                        'http': km_proxy_url,
                        'https': km_proxy_url,
                    }
                else:
                    self.data['keyManagerProxyRequestObject'] = {
                        'http': data['keyManagerProxyURL'],
                        'https': data['keyManagerProxyURL'],
                    }

            # datafeedVersion
            if "datafeedVersion" not in data:
                self.data["datafeedVersion"] = "v1"
            elif self.data.get("datafeedVersion").lower() == "v2":
                self.data["datafeedVersion"] = "v2"
            else:
                self.data["datafeedVersion"] = "v1"

            if 'datafeedEventsErrorTimeout' in data:
                self.data['datafeedEventsErrorTimeout'] = data['datafeedEventsErrorTimeout']

    def _build_url(self, host, port, contextPath):
        port = port if port else 443
        contextPath = contextPath if contextPath else ""

        if not (contextPath.startswith('/')):
            contextPath = '/' + contextPath
        if contextPath.endswith('/'):
            contextPath = contextPath[:-1]

        return 'https://' + host + ':' + str(port) + contextPath

    def get_agent_url(self):
        return self.data.get('agentUrl')

    def should_store_datafeed_id(self):
        return self.is_datafeed_v1() and self.is_datafeed_id_reused()

    def is_datafeed_v1(self):
        return DatafeedVersion.version_of(self.data.get("datafeedVersion")) != DatafeedVersion.V2

    def is_datafeed_id_reused(self):
        reuse_datafeed_id = self.data.get("reuseDatafeedID")
        return reuse_datafeed_id is None or reuse_datafeed_id

    def get_datafeed_id_folder_path(self):
        datafeed_id_file_path = self.data.get("datafeedIdFilePath")
        if datafeed_id_file_path:
            return datafeed_id_file_path
        return os.getcwd()



import os
import unittest
from unittest.mock import patch

from sym_api_client_python.clients.datafeed_client import DataFeedClient
from sym_api_client_python.clients.datafeed_client_v1 import DataFeedClientV1
from sym_api_client_python.clients.datafeed_client_v2 import DataFeedClientV2
from sym_api_client_python.configure.configure import SymConfig


class TestDataFeedClient(unittest.TestCase):
    def setUp(self):
        self.config = SymConfig(get_path_relative_to_resources_folder('./bot-config.json'))
        self.config.load_config()

    def test_datafeed_client_v1(self):
        self.assert_configured_df_version_leads_to_df_client_type('v1', DataFeedClientV1)

    def test_datafeed_client_v2(self):
        self.assert_configured_df_version_leads_to_df_client_type('v2', DataFeedClientV2)

    def test_datafeed_client_wrong_version(self):
        self.assert_configured_df_version_leads_to_df_client_type('25', DataFeedClientV1)

    def assert_configured_df_version_leads_to_df_client_type(self, datafeed_version, expected_class):
        with patch('sym_api_client_python.clients.sym_bot_client.SymBotClient') as mock_client:
            self.config.data['datafeedVersion'] = datafeed_version
            mock_client.get_sym_config.return_value = self.config
            datafeed_client = DataFeedClient(mock_client)

            self.assertIsInstance(datafeed_client.datafeed_client, expected_class)


def get_path_relative_to_resources_folder(path_relative_to_resources):
    path_to_resources = os.path.join(os.path.dirname(__file__), '../resources/', path_relative_to_resources)
    return os.path.normpath(path_to_resources)




from abc import ABC, abstractmethod

# Abstract class for message supression listener
# class is just an interface of functions to handle the connection events
# received from DataFeed Connection Event is "CONNECTIONREQUESTED"
# handle actual event logic in your implementation of
# this abstract class


class SuppressionListener(ABC):

    @abstractmethod
    def on_message_suppression(self, suppressed_message):
        pass


from yattag import Doc


class FormBuilder:
    """
    To start using Symphony Elements, you first need to create a form using the
    FormClient class.  The available elements can be found here: https://developers.symphony.com/symphony-developer/docs/available-components

    The following form contains setter functions corresponding to each element.

    To create a form, instantiate the FormBuilder class with a unique form id.  Add elements
    to this form by calling the appropriate setter function in the Form Class.

    Using this class allows a user to generate a messageML object corresponding to the elements inside the form.
    When this object is instantiated, it creates an empty string.  With every method called, the empty string is appended
    with messageML corresponding to the symphony element.  Lastly, users must call the format_element method in order to
    wrap the messageML string with the appropriate <messageML> tags.  At this point, the object can be sent as a message
    inside of Symphony.

    See examples/elementsExampleBot/form.py and examples/elementsExampleBot/message_processor.py for example usage:  
    """

    def __init__(self, form_id):
        self.form_id = form_id
        self.messageML = ''

    def format_element(self):
        """
        create_messageML() takes in a form object, parses its attritubes/metadata and returns
        a dictionary: {message : messageML object}.  This is the correct format to send
        in the Create Message v4 REST call.

        This messageML object is bootstrapped properly with <form> and
        <messageML> tags so that it can be rendered as an Element as is.
        """

        return dict(message='<messageML>' +
                            '<form id="{}">'.format(self.form_id)
                            + self.messageML +
                            '</form>'
                            '</messageML>')

    def add_header(self, text, size):
        """
        This function sets a header on the Form Element.
        Users must specify the text and size of the header that
        they wish to create.

        Returns a header: <h4>Form title</h4>
        """
        header = (size, text)
        doc, tag, text, line = Doc().ttl()
        with tag(header[0]):
            text(header[1])

        self.messageML += doc.getvalue()

    def add_button(self, name, display, type='action'):
        """
        This function adds a button to the Form Element.
        Users must specify the name, button text, and type.
        If no type is given, it's type will be set by default to 'action'

        Returns a button: <button name="test-button" type="action">Click Me!</button>
        """
        button = (name, type, display)
        doc, tag, text, line = Doc().ttl()
        with tag('button', name=name, type=type):
            text(display)

        self.messageML += doc.getvalue()

    def add_text_field(self, name, display, placeholder='', required='true', masked='true', maxlength=128, minlength=1):
        """
        This function adds a text field to a Form Element.
        Users must specifiy a name, display text, and/or optional parameters placeholder,
        required, maxlength, and minlength.

        Returns a text field: <text-field name="id1" placeholder="" required="true" maxlength="128" minlength="1">Input some text...</text-field>
        """
        doc, tag, text, line = Doc().ttl()

        with tag('text-field', name=name, placeholder=placeholder, required=required, masked=masked,
                 maxlength=maxlength, minlength=minlength):
            text(display)

        self.messageML += doc.getvalue()

    def add_text_area(self, name, display, placeholder='', required='true'):
        """
        This function adds a text area to a Form Element.
        Users must specifiy name, text and/or optional parameters, placeholder,
        and required.

        Returns a text area: <textarea name="id" placeholder="" required="true">My name is..</textarea>
        """
        doc, tag, text, line = Doc().ttl()
        with tag('textarea', name=name, placeholder=placeholder, required=required):
            text(display)

        self.messageML += doc.getvalue()

    def add_check_box(self, name, display, value='on', checked='false'):
        """
        This function adds a checkbox to a Form Element.
        Users must specify name, text, and/or optional parameters checked and value.
        Set checked='true' in order to have checkbox appear checked upon sending.

        Returns a checkbox: <checkbox name="id1" value="on" checked="false">Check Me!</checkbox>
        """
        doc, tag, text, line = Doc().ttl()

        with tag('checkbox', name=name, value=value, checked=checked):
            text(display)

        self.messageML += doc.getvalue()

    def add_radio_button(self, name, display, value='on', checked='false'):
        """
        This function adds a radio button to a Form Elment.
        Users must specify name, text, and/or optional parameters checked and value.
        Set checked='true' in order to have radio appear selected upon sending.

        Returns a radio button: <radio name="id1" value="on" checked="false">Click Me!</radio>
        """
        doc, tag, text, line = Doc().ttl()
        with tag('radio', name=name, value=value, checked=checked):
            text(display)

        self.messageML += doc.getvalue()

    def add_dropdown_menu(self, dropdown_group):
        """
        add_dropdown_menu_to_form([('dropdown', 'false', 'value1', 'false', 'Reeds Option'),
                                   ('dropdown', 'false', 'value2', 'false', 'Reeds Option2'),
                                   ('dropdown', 'false', 'value3', 'false', 'Reeds Option3')])

         --> <select name="dropdown" required="false">
               <option value="value1" selected="false">Reeds Option</option>
               <option value="value2" selected="false">Reeds Option2</option>
               <option value="value3" selected="false">Reeds Option3</option>
             </select>

        This function adds a dropdown menu to the Form Element.
        Users must specify a dropdown_group as a list of tuple's corresponding to each dropdown element.
        The dropdown group parameter must follow the following tuple pattern:

        (parent_name, required, value, selected, text)

        Users should pass a list of these tuples in order to create a dropdown group
        as shown above.
        """
        doc, tag, text, line = Doc().ttl()
        with tag('select', name=dropdown_group[0][0], required=dropdown_group[0][3]):
            for i in dropdown_group:
                with tag('option', value=i[2], selected=i[1]):
                    text(i[4])

        self.messageML += doc.getvalue()

    def add_person_selector(self, name, placeholder='', required='false'):
        """
        This function adds a person selector to the Form Element.
        Users must specify a name and/or optional parameters placeholder and required
        When a user is selected, the corresponding user_id is stored in an
        array corresponding to the name parameter

        Returns a Person Selector:  <person-selector name="awesome-users" placeholder="enter names..." required="false" />
        """
        doc, tag, text, line = Doc().ttl()
        doc.stag('person-selector', name=name, placeholder=placeholder, required=required)

        self.messageML += doc.getvalue()

    def add_table_selector(self, position, type, type_name, header_list, body_list, footer_list):
        """
        add_table_selector_to_form('left', 'checkbox', 'table-box',
                                    ['H1', 'H2', 'H3'],
                                    [["A1", "B1", "C1"],["A2", "B2", "C2"], ["A3", "B3", "C3"]],
                                    ["F1","F2","F3"])

        -->  <table>
              <thead>
                <tr>
                  <td>
                    <input type="checkbox" name="table-header" />
                  </td>
                  <td>H1</td>
                  <td>H2</td>
                  <td>H3</td>
                </tr>
              </thead>
              <tbody>
                <tr>
                  <td>
                    <input type="checkbox" name="table-box1" />
                  </td>
                  <td>A1</td>
                  <td>B1</td>
                  <td>C1</td>
                </tr>
                <tr>
                  <td>
                    <input type="checkbox" name="table-box2" />
                  </td>
                  <td>A2</td>
                  <td>B2</td>
                  <td>C2</td>
                </tr>
                <tr>
                  <td>
                    <input type="checkbox" name="table-box3" />
                  </td>
                  <td>A3</td>
                  <td>B3</td>
                  <td>C3</td>
                </tr>
              </tbody>
              <tfoot>
                <tr>
                  <td>
                    <input type="checkbox" name="table-footer" />
                  </td>
                  <td>F1</td>
                  <td>F2</td>
                  <td>F3</td>
                </tr>
              </tfoot>
            </table>

        This function adds a table selector to the Form Element.
        Users must specify a position, type, type_name, header_list, body_list, and footer list.

        position: 'left' or 'right',
        type: 'checkbox' or 'button'

        """
        doc, tag, text, line = Doc().ttl()
        with tag('table'):
            with tag('thead'):
                with tag('tr'):
                    if position == 'left':
                        if type == 'button':
                            line('td', 'Select')
                            for i in table_selector:
                                for j in header_list:
                                    line('td', j)
                        else:
                            with tag('td'):
                                doc.input(name='table-header', type='checkbox')
                                for j in header_list:
                                    line('td', j)
                    else:
                        if type == 'button':
                            for j in header_list:
                                line('td', j)
                            line('td', 'Select')

                        else:
                            for i in table_selector:
                                for j in header_list:
                                    line('td', j)
                            with tag('td'):
                                doc.input(name='table-header', type='checkbox')

            with tag('tbody'):
                for num, j in enumerate(body_list, start=1):
                    with tag('tr'):
                        if position == 'left':
                            with tag('td'):
                                if type == 'button':
                                    with tag('button', name=type_name + str(num)):
                                        text('Button')
                                else:
                                    doc.input(name=type_name + str(num), type='checkbox')
                            for k in j:
                                line('td', k)
                        else:
                            for k in j:
                                line('td', k)
                            with tag('td'):
                                if type == 'button':
                                    with tag('button', name=type_name + str(num)):
                                        text('Button')
                                else:
                                    doc.input(name=type_name + str(num), type='checkbox')

            with tag('tfoot'):
                with tag('tr'):
                    if position == 'left':
                        with tag('td'):
                            if type == 'button':
                                with tag('button', name="footer"):
                                    text('Button')
                            else:
                                doc.input(name='table-footer', type='checkbox')
                        for j in footer_list:
                            line('td', j)
                    else:
                        for j in footer_list:
                            line('td', j)
                        with tag('td'):
                            if type == 'button':
                                with tag('button', name="footer"):
                                    text('Button')
                            else:
                                doc.input(name='table-footer', type='checkbox')

        self.messageML += doc.getvalue()



from .wall_post_listener import WallPostListener
import logging
# A sample implementation of Abstract WallPostListener class
# The listener can respond to incoming events if the respective event
# handler has been implemented


class WallPostListenerImp(WallPostListener):
    """Example implementation of WallPostListener

        sym_bot_client: contains clients which respond to incoming events

    """

    def __init__(self, sym_bot_client):
        self.bot_client = sym_bot_client

    def on_wall_post_msg(self, wall_post_msg):
        logging.debug('receieved incoming wall post %s', wall_post_msg)

    def on_shared_post(self, shared_post):
        logging.debug('received incoming shared post %s', shared_post)



import logging

from .api_client import APIClient


# child class of APIClient --> Extends error handling functionality
# AdminClient class contains a series of functions corresponding to all
# pod admin endpoints on the REST API.
class AdminClient(APIClient):

    def __init__(self, bot_client):
        self.bot_client = bot_client

    def admin_get_user(self, user_id):
        """
        Returns details for a particular user.

        Calling this endpoint requires the
        ACCESS_USER_PROVISIONING_API and ACCESS_ADMIN_API privileges.
        """
        logging.debug('AdminClient/admin_get_user()')
        url = '/pod/v2/admin/user/{0}'.format(user_id)
        return self.bot_client.execute_rest_call("GET", url)

    def admin_list_users(self, skip=0, limit=50):
        """
        Returns a list of users ID, including user metadata

        Calling this endpoint requires the ACCESS_USER_PROVISIONING_API privilege.
        """
        logging.debug('AdminClient/admin_list_users()')
        url = '/pod/v2/admin/user/list'
        params = {'skip': skip, 'limit': limit}

        return self.bot_client.execute_rest_call("GET", url, params=params)
    
    def admin_create_user(self, user_attributes):
        """
        Creates a new user, either End-User or Service User.

        --End-User Accounts are assigned to employees. 
        To create an end user account, the accountType field must be NORMAL.
        -- Service User Accounts are a type of account used for bots or applications, 
        rather than end-users. To create a service user account, 
        the accountType field must be SYSTEM.

        Calling this endpoint requires the
        ACCESS_USER_PROVISIONING_API and ACCESS_ADMIN_API privileges.

        TODO: Add `Password` object generation
        """
        logging.debug('AdminClient/admin_list_users()')
        url = '/pod/v2/admin/user/create'
        return self.bot_client.execute_rest_call("POST", url, json=user_attributes)
    
    def admin_update_user(self, user_id, updated_user_attributes):
        """
        Updates an existing user.

        BODY PARAMS

            emailAddress
            The user's email address.

            firstName
            The user's first name.

            lastName
            The user's last name.

            userName
            The user's name. The username must be unique and must not match any existing usernames or app IDs.

            displayName
            The user's display name.

            companyName
            The name of the user's company.

            department
            The user's department.

            division
            The user's division.

            title
            The user's title.

            twoFactorAuthPhone
            The user's two factor authentication mobile phone number.

            workPhoneNumber
            The user's work phone number.

            mobilePhoneNumber
            The user's mobile number.

            smsNumber
            The user's SMS number.

            accountType
            The user's account type. Possible values: NORMAL, SYSTEM.

            location
            The user's location.

            jobFunction
            The user's job function. Possible values: Project Manager, Trader, Sales, Strategist, Business Development Executive, Corporate Access, Analyst, Other, Research Analyst, Developer, Economist, Portfolio Manager, Director.

            assetClasses
            The user's asset classes. Possible values: Fixed Income, Currencies, Commodities, Equities.

            industries
            The user's industries. Possible values: Conglomerates, Healthcare, Transportation, Services, Energy & Utilities, Real Estate, Consumer Cyclicals, Financials, Basic Materials, Consumer Non-Cyclicals, Technology.

            currentKey
            A UserKeyRequest Object object containing the current RSA key information to use for authenticating the user. When updating the user, this can be set (rotated) to a new key, in which case the existing key is stored in the previousKey and can be used for authentication during the current session. For more information see RSA Bot Authentication Workflow.

            previousKey
            A UserKeyRequest Object object containing the RSA key information from the previous key in a key rotation scenario. The previous key can be used when the currentKey has been updated during the current session. The previousKey is valid for 72 hours. For more information see RSA Bot Authentication Workflow.



        Calling this endpoint requires the ACCESS_USER_PROVISIONING_API and ACCESS_ADMIN_API privileges.
        """
        logging.debug('AdminClient/admin_update_user()')
        url = '/pod/v2/admin/user/{0}/update'.format(user_id)
        return self.bot_client.execute_rest_call("POST", url, json=updated_user_attributes)

    def admin_get_user_avatar(self, user_id):
        """
        Returns the URL of the avatar of a particular user.

        Calling this endpoint requires the ACCESS_USER_PROVISIONING_API privilege.
        """
        logging.debug('AdminClient/admin_get_user_avatar')
        url = '/pod/v1/admin/user/{0}/avatar'.format(user_id)
        return self.bot_client.execute_rest_call("GET", url)

    def admin_update_avatar(self, user_id, image_encoded_string):
        """
        Updates the avatar of a particular user. file_path to base64 encoded image

        TODO: base64 encode the image in the file_path

        Calling this endpoint requires the User Provisioning role with ACCESS_USER_PROVISIONING_API privilege.
        """
        logging.debug('AdminClient/admin_update_avatar')
        url = '/pod/v1/admin/user/{0}/avatar/update'.format(user_id)
        #base64encode the image file
        data = {'image': image_encoded_string}
        return self.bot_client.execute_rest_call("POST", url, json=data)    

    def admin_get_user_status(self, user_id):
        """
        Get the status, active or inactive, for a particular user.

        Calling this endpoint requires the User Provisioning role with ACCESS_USER_PROVISIONING_API privilege.
        """
        logging.debug('AdminClient/admin_get_user_status()')
        url = '/pod/v1/admin/user/{0}/status'.format(user_id)
        return self.bot_client.execute_rest_call("GET", url)

    def admin_update_user_status(self, user_id, status):
        """
        Update the status of a particular user.

        `status` can be 'ENABLED' or 'DISABLED' 

        Calling this endpoint requires the User Provisioning role with ACCESS_USER_PROVISIONING_API privilege.
        """
        logging.debug('AdminClient/admin_update_user_status()')
        url = '/pod/v1/admin/user/{0}/status/update'.format(user_id)
        data  = {'status': status}
        return self.bot_client.execute_rest_call("POST", url, json=data)

    def admin_list_pod_features(self):
        """
        Returns the full set of Symphony features available for this pod.

        Features entitlements definition:

            postReadEnabled: Allows the user to read wall posts.
            postWriteEnabled: Allows the user to write wall posts.
            delegatesEnabled: Allows the user to have delegates.
            isExternalIMEnabled: Allows the user to chat in external IM/MIMs.
            canShareFilesExternally: Allows the user to share files externally.
            canCreatePublicRoom: Allows the user to create internal public rooms.
            canUpdateAvatar: Allows the user to edit profile picture.
            isExternalRoomEnabled: Allows the user to chat in private external rooms.
            canCreatePushedSignals: Allows the user to create push signals.
            canUseCompactMode: Enables Lite Mode.
            mustBeRecorded: Must be recorded in meetings.
            sendFilesEnabled: Allows the user to send files internally.
            canUseInternalAudio: Allows the user to use audio in internal Meetings.
            canUseInternalVideo: Allows the user to use video in internal Meetings.
            canProjectInternalScreenShare: Allows the user to share screens in internal Meetings.
            canViewInternalScreenShare: Allows the user to view shared screens in internal Meetings.
            canCreateMultiLateralRoom: Allows the user to create multi-lateral room.
            canJoinMultiLateralRoom: Allows the user to join multi-lateral room.
            canUseFirehose: Allows the user to use Firehose.
            canUseInternalAudioMobile: Allows the user to use audio in internal meetings on mobile.
            canUseInternalVideoMobile: Allows the user to use video in internal meetings on mobile.
            canProjectInternalScreenShareMobile: Allows the user to share screens in internal meetings on mobile.
            canViewInternalScreenShareMobile: Allows the user to view shared screens in internal meetings on mobile.
            canManageSignalSubscription: Allows the user to manage signal subscriptions.

Calling this endpoint requires the User Provisioning role with ACCESS_USER_PROVISIONING_API privilege.
"""
        logging.debug('AdminClient/admin_list_pod_features()')
        url = '/pod/v1/admin/system/features/list'
        return self.bot_client.execute_rest_call("GET", url)

    def admin_get_user_features(self, user_id):
        """
        Returns the list of Symphony feature entitlements for a particular user.

        Use the data returned from this endpoint with Find Users to filter users by a specific entitlement.

        Calling this endpoint requires the User Provisioning role with ACCESS_USER_PROVISIONING_API privilege.
        """
        logging.debug('AdminClient/admin_get_user_features()')
        url = '/pod/v1/admin/user/{0}/features'.format(user_id)
        return self.bot_client.execute_rest_call("GET", url)

    def admin_update_user_features(self, user_id, feature_list):
        """
        Updates the feature entitlements for a particular user.

        featureList -the features to update. Specified by entitlement name and enabled (true or false).

        Calling this endpoint requires the User Provisioning role with ACCESS_USER_PROVISIONING_API privilege.
        """
        logging.debug('AdminClient/admin_update_user_features()')
        url = '/pod/v1/admin/user/{0}/features/update'.format(user_id)
        return self.bot_client.execute_rest_call("POST", url, json=feature_list)

    def admin_find_users(self, filters, skip=0, limit=50):
        """
        Finds a list of users based on a specified role or feature entitlement.

        'filters' is a dictionary that can contain:
            'role'    -- int64: User role
            'feature' -- string: Feature entitlement value
            'status'  -- string: 'ENABLED' or 'DISABLED'

        Calling this endpoint requires the User Provisioning role
        with ACCESS_USER_PROVISIONING_API privilege.
        """ 
        logging.debug('AdminClient/admin_find_users()')
        url = '/pod/v1/admin/user/find'
        params = {'skip': skip, 'limit': limit}
        return self.bot_client.execute_rest_call("POST", url, params=params, json=filters)

    def admin_list_roles(self):
        """
        Returns a list of all roles available in the company (pod).

        Calling this endpoint requires the ACCESS_ADMIN_API privilege.
        """
        logging.debug('AdminClient/admin_list_roles()')
        url = '/pod/v1/admin/system/roles/list'
        return self.bot_client.execute_rest_call("GET", url)

    def admin_add_role(self, user_id, payload={}):
        """
        Add a role or optional entitleable action to a user's account. For example: {"id":"COMPLIANCE_OFFICER.MONITOR_ROOMS"}

        Calling this endpoint requires the User Provisioning role with ACCESS_USER_PROVISIONING_API privilege.
        """
        logging.debug('AdminClient/admin_add_role()')
        url = '/pod/v1/admin/user/{0}/roles/add'.format(user_id)

        return self.bot_client.execute_rest_call("POST", url, json=payload)

    def admin_remove_role(self, user_id, payload={}):
        """
        Remove a role or optional entitleable action to a user's account. For example: {"id":"L2_SUPPORT"}

        Calling this endpoint requires the User Provisioning role with ACCESS_USER_PROVISIONING_API privilege.
        """
        logging.debug('AdminClient/admin_remove_role()')
        url = '/pod/v1/admin/user/{0}/roles/remove'.format(user_id)
        return self.bot_client.execute_rest_call("POST", url, json=payload)


    def import_message(self, imported_message):
        logging.debug('MessageClient/import_message()')
        url = '/agent/v4/message/import'
        return self.bot_client.execute_rest_call("POST", url, json=imported_message)

    # go on admin clients
    def suppress_message(self, message_id):
        logging.debug('MessageClient/suppress_message()')
        url = '/pod/v1/admin/messagesuppression/{0}/suppress'.format(message_id)
        return self.bot_client.execute_rest_call("POST", url)






from abc import ABC, abstractmethod
import logging

from .datafeed_id_repository import OnDiskDatafeedIdRepository
from ..listeners.elements_listener import ElementsActionListener
from ..listeners.connection_listener import ConnectionListener
from ..listeners.im_listener import IMListener
from ..listeners.room_listener import RoomListener

from ..auth.auth_endpoint_constants import auth_endpoint_constants

log = logging.getLogger(__name__)


class AbstractDatafeedEventService(ABC):

    def __init__(self, sym_bot_client, error_timeout_sec=None, maximum_timeout_sec=None):
        self.datafeed_events = []
        self.room_listeners = []
        self.im_listeners = []
        self.connection_listeners = []
        self.elements_listeners = []
        self.suppression_listeners = []
        self.wall_post_listeners = []
        self.registered_triggers = []

        self.bot_client = sym_bot_client
        self.config = sym_bot_client.get_sym_config()
        self.datafeed_client = self.bot_client.get_datafeed_client()
        self.datafeed_id_repository = OnDiskDatafeedIdRepository(self.config.get_datafeed_id_folder_path())
        self.stop = False

        # TODO: Should not be handled in the DF ES like this, why put a timeout in the config if we can put a parameter with default value(config related)
        # Timeout will start at and eventually reset to this
        _config_key = 'datafeedEventsErrorTimeout'
        # Mapping of Event type to the function to handle it
        self.routing_dict = {
            'MESSAGESENT': self.msg_sent_handler,
            'MESSAGESUPPRESSED': self.suppressed_message_handler,
            'INSTANTMESSAGECREATED': self.instant_msg_handler,
            'ROOMCREATED': self.room_created_handler,
            'ROOMDEACTIVATED': self.room_deactivated_handler,
            'ROOMREACTIVATED': self.room_reactivated_handler,
            'ROOMUPDATED': self.room_updated_handler,
            'USERJOINEDROOM': self.user_joined_room_handler,
            'USERLEFTROOM': self.user_left_room_handler,
            'ROOMMEMBERPROMOTEDTOOWNER': self.promoted_to_owner,
            'ROOMMEMBERDEMOTEDFROMOWNER': self.demoted_from_owner,
            'CONNECTIONACCEPTED': self.connection_accepted_handler,
            'CONNECTIONREQUESTED': self.connection_requested_handler,
            'SYMPHONYELEMENTSACTION': self.elements_action_handler,
            'SHAREDPOST': self.shared_post_handler,
        }

        if error_timeout_sec is None:
            self.baseline_timeout_sec = self.config.data.get(_config_key, auth_endpoint_constants["TIMEOUT"])
        else:
            if _config_key in self.config.data:
                log.debug('{} listed in config, but overriden to {}s by parammeter'
                         .format(_config_key, error_timeout_sec))
            self.baseline_timeout_sec = error_timeout_sec

        self.current_timeout_sec = self.baseline_timeout_sec
        # Never wait less than this
        self.lower_threshold = self.baseline_timeout_sec
        # Raise a RuntimeError once this upper threshold exceeded
        if maximum_timeout_sec is None:
            self.upper_threshold = self.config.data.get('datafeedEventsErrorMaxTimeout', 60)
        else:
            self.upper_threshold = maximum_timeout_sec
        # After every failure multiply the timeout by a factor
        self.timeout_multiplier = 2

    @abstractmethod
    def start_datafeed(self):
        pass

    def activate_datafeed(self):
        if self.stop:
            self.stop = False

    def deactivate_datafeed(self):
        if not self.stop:
            self.stop = True

    ### Listeners ###
    def add_listeners(self, *listeners):
        for listener in listeners:
            if isinstance(listener, ElementsActionListener):
                self.add_elements_listener(listener)
            elif isinstance(listener, ConnectionListener):
                self.add_connection_listener(listener)
            elif isinstance(listener, IMListener):
                self.add_im_listener(listener)
            elif isinstance(listener, RoomListener):
                self.add_room_listener(listener)

    def remove_listeners(self, *listeners):
        for listener in listeners:
            if isinstance(listener, ElementsActionListener):
                self.remove_elements_listener(listener)
            elif isinstance(listener, ConnectionListener):
                self.remove_connection_listener(listener)
            elif isinstance(listener, IMListener):
                self.remove_im_listener(listener)
            elif isinstance(listener, RoomListener):
                self.remove_room_listener(listener)

    def add_room_listener(self, room_listener):
        self.room_listeners.append(room_listener)

    def remove_room_listener(self, room_listener):
        self.room_listeners.remove(room_listener)

    def add_im_listener(self, im_listener):
        self.im_listeners.append(im_listener)

    def remove_im_listener(self, im_listener):
        self.im_listeners.remove(im_listener)

    def add_connection_listener(self, connection_listener):
        self.connection_listeners.append(connection_listener)

    def remove_connection_listener(self, connection_listener):
        self.connection_listeners.remove(connection_listener)

    def add_elements_listener(self, elements_listener):
        self.elements_listeners.append(elements_listener)

    def remove_elements_listener(self, elements_listener):
        self.elements_listeners.remove(elements_listener)

    def add_wall_post_listener(self, wall_post_listener):
        self.wall_post_listeners.append(wall_post_listener)

    def remove_wall_post_listener(self, wall_post_listener):
        self.wall_post_listeners.remove(wall_post_listener)

    def add_suppression_listener(self, suppression_listener):
        self.suppression_listeners.append(suppression_listener)

    def remove_suppression_listener(self, suppression_listener):
        self.suppression_listeners.remove(suppression_listener)


    # TODO: Add doc
    # TODO: Change all brackets access types by get function
    def handle_events(self, events):
        """ Routes the event to the proper handler.

            Also Checks to see that the sender's email is not equal to Bot's email in config file.
            This functionality helps the bot avoid entering an infinite loop where it responds to its own messageSent
            after an event comes back over the dataFeed
        """
        log.debug('DataFeedEventService/handle_events()')
        for event in events:
            if event is None:
                continue
            
            log.debug(
                'DataFeedEventService/read_datafeed() --> '
                'Incoming event with id: {}'.format(event.get('id'))
            )
            
            if event['initiator']['user']['userId'] == self.bot_client.get_bot_user_info()['id']:
                continue
            else:
                self.handle_event(event)

    # function takes in single event --> Checks eventType --> forwards event
    # to proper handling function there is a handle_event function that
    # corresponds to each eventType
    def handle_event(self, payload):
        log.debug('DataFeedEventService/handle_event()')
        event_type = str(payload['type'])
        try:
            route = self.routing_dict[event_type]
        except KeyError:
            log.error('Unrecognised event type: ' + event_type)
            return

        route(payload)

    ### Handlers ###
    def msg_sent_handler(self, payload):
        """This handler is used for both room messages and IMs. Which listener is invoked
        depends on the streamType"""
        log.debug('msg_sent_handler function started')
        stream_type = payload['payload']['messageSent']['message']['stream']['streamType']
        message_sent_data = payload['payload']['messageSent']['message']
        if str(stream_type) == 'ROOM':
            for listener in self.room_listeners:
                listener.on_room_msg(message_sent_data)
        elif str(stream_type) == 'POST':
            for listener in self.wall_post_listeners:
                listener.on_wall_post_msg(message_sent_data)
        else:
            for listener in self.im_listeners:
                listener.on_im_message(message_sent_data)

    def instant_msg_handler(self, payload):
        log.debug('instant_msg_handler function started')
        instant_message_data = payload['payload']['instantMessageCreated']
        for listener in self.im_listeners:
            listener.on_im_created(instant_message_data)

    def room_created_handler(self, payload):
        log.debug('room_created_handler function started')
        room_created_data = payload['payload']['roomCreated']
        for listener in self.room_listeners:
            listener.on_room_created(room_created_data)

    def room_updated_handler(self, payload):
        log.debug('room_updated_handler')
        room_updated_data = payload['payload']['roomUpdated']
        for listener in self.room_listeners:
            listener.on_room_updated(room_updated_data)

    def room_deactivated_handler(self, payload):
        log.debug('room_deactivated_handler')
        room_deactivated_data = payload['payload']['roomDeactivated']
        for listener in self.room_listeners:
            listener.on_room_deactivated(room_deactivated_data)

    def room_reactivated_handler(self, payload):
        log.debug('room_reactivated_handler')
        room_reactivated_data = payload['payload']['roomReactivated']
        for listener in self.room_listeners:
            listener.on_room_reactivated(room_reactivated_data)

    def user_joined_room_handler(self, payload):
        log.debug('user_joined_room_handler')
        user_joined_room_data = payload['payload']['userJoinedRoom']
        for listener in self.room_listeners:
            listener.on_user_joined_room(user_joined_room_data)

    def user_left_room_handler(self, payload):
        log.debug('user_left_room_handler')
        user_left_room_data = payload['payload']['userLeftRoom']
        for listener in self.room_listeners:
            listener.on_user_left_room(user_left_room_data)

    def promoted_to_owner(self, payload):
        log.debug('promoted_to_owner')
        promoted_to_owner_data = payload['payload']['roomMemberPromotedToOwner']
        for listener in self.room_listeners:
            listener.on_room_member_promoted_to_owner(promoted_to_owner_data)

    def demoted_from_owner(self, payload):
        log.debug('demoted_from_owner')
        demoted_from_owner_data = payload['payload']['roomMemberDemotedFromOwner']
        for listener in self.room_listeners:
            listener.on_room_member_demoted_from_owner(demoted_from_owner_data)

    def connection_accepted_handler(self, payload):
        log.debug('connection_accepted_handler')
        connection_accepted_data = payload['payload']['connectionAccepted']
        for listener in self.connection_listeners:
            listener.on_connection_accepted(connection_accepted_data)

    def connection_requested_handler(self, payload):
        log.debug('connection_requested_handler')
        connection_requested_data = payload['payload']['connectionRequested']
        for listener in self.connection_listeners:
            listener.on_connection_requested(connection_requested_data)

    def elements_action_handler(self, payload):
        log.debug('elements_action_handler')
        for listener in self.elements_listeners:
            listener.on_elements_action(payload)

    def shared_post_handler(self, payload):
        log.debug('shared_post_handler')
        shared_post = payload['payload']['sharedPost']
        for listener in self.wall_post_listeners:
            listener.on_shared_post(shared_post)

    def suppressed_message_handler(self, payload):
        log.debug('suppressed_message_handler')
        message_suppressed = payload['payload']['messageSuppressed']
        for listener in self.suppression_listeners:
            listener.on_message_suppression(message_suppressed)

    ### Handle errors ###
    def get_and_increase_timeout(self, previous_exc=None):
        """Return the current timeout and then increase it for next time. Throw RuntimeError if the
        current timeout is larger than the upper threshold - if one exists.

            previous_exc: provide the exception that causes the original timeout to have that
                          attached to the stacktrace of the runtime error, showing more clearly
                          why the timeout occurred. """

        if (self.upper_threshold is not None) and (self.current_timeout_sec > self.upper_threshold):
            e = RuntimeError(
                'Upper timeout threshold exceeded: {:.4g}s'.format(self.upper_threshold)
            )
            if previous_exc is not None:
                raise e from previous_exc
            else:
                raise e
        else:
            original = self.current_timeout_sec
            new_timeout = original * self.timeout_multiplier
            self.current_timeout_sec = new_timeout
            log.debug('DataFeedEventService/get_and_increase_timeout --> '
                      'Using current timeout of {:.4g}s, but increasing to {:.4g}s for next time'
                      .format(original, new_timeout))
            return max(self.lower_threshold, original)

    def decrease_timeout(self):
        original = self.current_timeout_sec
        new_timeout = self.baseline_timeout_sec
        if original != new_timeout:
            log.debug('DataFeedEventService/decrease_timeout() --> '
                      'Decreasing timeout from {:.4g}s to {:.4g}s'.format(original, new_timeout))
            self.current_timeout_sec = new_timeout
        return self.current_timeout_sec

    def _get_from_file_or_create_datafeed_id(self):
        if self.config.should_store_datafeed_id():
            datafeed_id = self.datafeed_id_repository.read_datafeed_id_from_file()
            if datafeed_id:
                return datafeed_id
        return self._create_datafeed_and_persist()

    def _create_datafeed_and_persist(self):
        datafeed_id = self.datafeed_client.create_datafeed()
        if self.config.should_store_datafeed_id():
            self.datafeed_id_repository.store_datafeed_id_to_file(datafeed_id, self.config.get_agent_url())
        return datafeed_id



import asyncio
from unittest import mock
from unittest.async_case import IsolatedAsyncioTestCase
from unittest.mock import MagicMock, AsyncMock

from sym_api_client_python.clients.sym_bot_client import SymBotClient
from sym_api_client_python.configure.configure import SymConfig
from sym_api_client_python.datafeed_event_service import AsyncDataFeedEventService
from sym_api_client_python.listeners.im_listener import IMListener
from tests.clients.test_datafeed_client import get_path_relative_to_resources_folder


class TestDataFeedEventService(IsolatedAsyncioTestCase):

    def setUp(self):
        self.config = SymConfig(get_path_relative_to_resources_folder('./bot-config.json'))
        self.config.load_config()
        self.client = SymBotClient(None, self.config)
        self.ran = False

    @mock.patch(
        'sym_api_client_python.clients.datafeed_client.DataFeedClient',
        new_callable=AsyncMock)
    async def test_read_datafeed_event_no_id(self, datafeed_client_mock):
        service = AsyncDataFeedEventService(self.client)
        self.client.get_bot_user_info = MagicMock(return_value={'id': 456})

        service.datafeed_client = datafeed_client_mock
        datafeed_client_mock.read_datafeed_async.side_effect = self.return_event_no_id_first_time

        listener = IMListenerRecorder(service)
        service.add_im_listener(listener)

        # Simulate start_datafeed
        await asyncio.gather(service.read_datafeed(), service.handle_events())

        self.assertIsNotNone(listener.last_message)

    async def return_event_no_id_first_time(self, _arg):
        if self.ran:
            # Give control back to handle_event coroutine
            await asyncio.sleep(0)
            return []

        self.ran = True
        event = {'type': 'MESSAGESENT', 'timestamp': 0,
                 'payload': {'messageSent': {'message': {'stream': {'streamType': 'IM'}}}},
                 'initiator': {'user': {'userId': 123}}}
        return [event]


class IMListenerRecorder(IMListener):

    def __init__(self, service) -> None:
        super().__init__()
        self.service = service
        self.last_message = None

    async def on_im_message(self, message):
        self.last_message = message
        # Stop datafeed loop
        await self.service.deactivate_datafeed(False)

    def on_im_created(self, stream):
        pass  # Not used



import logging

from .api_client import APIClient
from .datafeed_client_v1 import DataFeedClientV1
from .datafeed_client_v2 import DataFeedClientV2


# child class of APIClient --> Extends error handling functionality
class DataFeedClient(APIClient):

    def __init__(self, bot_client):
        self.config = bot_client.get_sym_config()

        if self.config.is_datafeed_v1():
            self.datafeed_client = DataFeedClientV1(bot_client)
        else:
            self.datafeed_client = DataFeedClientV2(bot_client)

    # raw api call to create_datafeed --> returns datafeed_id
    def create_datafeed(self):
        """
        Create a new real time messages / events stream ("datafeed").
        The datafeed provides messages and events from all conversations that the user is in.
        The types of events surfaced in the datafeed can be found in the Real Time Events list.

        Returns the ID of the datafeed that has just been created. This ID should then be used as input to the Read Messages/Events Stream v4 endpoint.

        The datafeed will expire if it is not read before its capacity is reached.

        There is a maximum of 5 Datafeed per Service Account configured by default.
        This setting can be changed in the API Agent's property file.
        """
        return self.datafeed_client.create_datafeed()

    # raw api call to read_datafeed --> returns an array of events returned
    # from DataFeed
    def read_datafeed(self, datafeed_id, *ackId):
        """
        Reads messages from a given real time messages / events stream ("datafeed").

        The datafeed provides messages and events from all conversations that the user is in.
        The types of events surfaced in the datafeed can be found in the Real Time Events list.

        If no more messages are available, this method will be blocked for 30 seconds and return an HTTP 204 response (No Content) after. It is intended that the client should re-invoke this method as soon as it has processed the messages received in the previous call.
        If the client is able to consume messages more quickly than they become available, each call will be initially blocked and there is no need to delay before re-invoking this method.

        A datafeed will expire if its unread capacity is reached. For a standard datafeed, this will be 250 queued messages, and for firehose 500 messages.

        A datafeed can only be consumed by one client thread at a time. For example, polling the datafeed by two threads may lead to messages being delivered out of order.

        It is mandatory to read the datafeed from the same Agent instance that it was created on, bypassing any load balancer that may be in front of a multi-Agent configuration.

        The caller needs to be the same user that created the datafeed.

        There is no guarantee of message delivery for this endpoint. If errors occur while processing the request, either on the server or the API Agent, the messages that would have been returned may not be readable from the current stream. The messages are still available via the Messages endpoint and the Content Export function.

        """
        return self.datafeed_client.read_datafeed(datafeed_id, *ackId)

    def list_datafeed_id(self):
        """
        List datafeeds for a user's auth session.

        Returns
        -------
        Dictionary of datafeed ids a user is following. {"id": datafeed_id1, "id": datafeed_id2 ...}

        This feature is not supported in datafeed v1.
        """
        return self.datafeed_client.list_datafeed_id()

    def delete_datafeed(self, datafeed_id):
        """
        Delete a given datafeed

        Parameters
        ----------
        datafeed_id : int
            Datafeed id of the datafeed we want to delete.

        This feature is not supported in datafeed v1.
        """
        self.datafeed_client.delete_datafeed(datafeed_id)

    def get_ack_id(self):
        """
        Get the ack id for reading a V2 datafeed.

        This is stored in the DF client from the previous datafeed read.
        We give the user the freedom of altering the handling of the ack Id.

        Returns
        -------
        Ack id present in the previous read. Empty string if no read has been done

        This feature is not supported in datafeed v1.
        """
        return self.datafeed_client.get_ack_id()


    async def read_datafeed_async(self, datafeed_id):
        """
        This works the same as the previous datafeed apart from it's asynchronous and therefore should be called with the await keyword

        See read_datafeed for more info
        """
        return await self.datafeed_client.read_datafeed_async(datafeed_id)



import logging
import unittest

from sym_api_client_python.processors.sym_message_parser import SymMessageParser


class TestMessageParser(unittest.TestCase):

    def setUp(self):
        logging.debug('testing Message Parser File:')
        self.message_parser = SymMessageParser()
        self.test_message_json = {
                            "messageId": "LKX55AtesGPxrv16ed_TIX___pMru4hmbQ",
                            "timestamp": 1566929352601,
                            "message": "<div data-format=\"PresentationML\" data-version=\"2.0\" class=\"wysiwyg\"><p><span class=\"entity\" data-entity-id=\"0\">$aapl</span> <span class=\"entity\" data-entity-id=\"1\">$amzn</span> <span class=\"entity\" data-entity-id=\"2\">$nasdaq</span></p></div>",
                            "data": "{\"0\":{\"id\":[{\"type\":\"org.symphonyoss.fin.security.id.ticker\",\"value\":\"aapl\"}],\"type\":\"org.symphonyoss.fin.security\",\"version\":\"1.0\"},\"1\":{\"id\":[{\"type\":\"org.symphonyoss.fin.security.id.ticker\",\"value\":\"amzn\"}],\"type\":\"org.symphonyoss.fin.security\",\"version\":\"1.0\"},\"2\":{\"id\":[{\"type\":\"org.symphonyoss.fin.security.id.ticker\",\"value\":\"nasdaq\"}],\"type\":\"org.symphonyoss.fin.security\",\"version\":\"1.0\"}}",
                            "user": {
                                "userId": 349026222344902,
                                "firstName": "Reed",
                                "lastName": "Feldman",
                                "displayName": "Reed Feldman (Develop 2)",
                                "email": "reed.feldman@symphony.com",
                                "username": "reed.feldman@symphony.com"
                            },
                            "stream": {
                                "streamId": "sOKpwRk_5_N838P10ATuFX___pNk9zJndA",
                                "streamType": "ROOM"
                            },
                            "externalRecipients": "false",
                            "userAgent": "DESKTOP-40.0.0-10665-MacOSX-10.14.6-Chrome-76.0.3809.100",
                            "originalFormat": "com.symphony.messageml.v2"
                        }

    def test_get_text(self):
        print(self.message_parser.get_text(self.test_message_json))

    def test_get_im_firstname(self):
        print(self.message_parser.get_im_first_name(self.test_message_json))

    def test_get_stream_id(self):
        print(self.message_parser.get_stream_id(self.test_message_json))

    def test_get_mentions(self):
        print(self.message_parser.get_mentions(self.test_message_json))

    def test_get_mention_ids(self):
        print(self.message_parser.get_mention_ids(self.test_message_json))

    def test_get_hash_tags(self):
        print(self.message_parser.get_hash_tags(self.test_message_json))

    def test_get_hash_tag_values(self):
        print(self.message_parser.get_hash_tag_values(self.test_message_json))

    def test_cash_tags(self):
        print(self.message_parser.get_cash_tags(self.test_message_json))

    def test_get_cash_tag_values(self):
        print(self.message_parser.get_cash_tag_values(self.test_message_json))


if __name__ == '__main__':
    unittest.main()
